/*
 * Name:    d_epmcs.c 
 *
 * Purpose: MX multichannel scaler driver for the EPICS multichannel
 *          scaler support.  This currently is just for the Struck MCS.
 *
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 *
 *          Please note that the EPICS MCS record numbers scalers starting
 *          at 1, but the mcs->data_array numbers scalers starting at 0.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_meas.h"
#include "mx_mcs.h"
#include "d_epmcs.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the mcs driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_mcs_record_function_list = {
	mxd_epics_mcs_initialize_type,
	mxd_epics_mcs_create_record_structures,
	mxd_epics_mcs_finish_record_initialization,
	mxd_epics_mcs_delete_record,
	NULL,
	mxd_epics_mcs_read_parms_from_hardware,
	mxd_epics_mcs_write_parms_to_hardware,
	mxd_epics_mcs_open,
	mxd_epics_mcs_close
};

MX_MCS_FUNCTION_LIST mxd_epics_mcs_mcs_function_list = {
	mxd_epics_mcs_start,
	mxd_epics_mcs_stop,
	mxd_epics_mcs_clear,
	mxd_epics_mcs_busy,
	mxd_epics_mcs_read_all,
	mxd_epics_mcs_read_scaler,
	mxd_epics_mcs_read_measurement,
	mxd_epics_mcs_get_parameter,
	mxd_epics_mcs_set_parameter
};

/* EPICS mcs data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_mcs_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCS_STANDARD_FIELDS,
	MXD_EPICS_MCS_STANDARD_FIELDS
};

long mxd_epics_mcs_num_record_fields
		= sizeof( mxd_epics_mcs_record_field_defaults )
			/ sizeof( mxd_epics_mcs_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_mcs_rfield_def_ptr
			= &mxd_epics_mcs_record_field_defaults[0];

static mx_status_type mxd_epics_mcs_configure(
				MX_MCS *mcs, MX_EPICS_MCS *epics_mcs );

/* A private function for the use of the driver. */

static mx_status_type
mxd_epics_mcs_get_pointers( MX_MCS *mcs,
			MX_EPICS_MCS **epics_mcs,
			const char *calling_fname )
{
	const char fname[] = "mxd_epics_mcs_get_pointers()";

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( epics_mcs == (MX_EPICS_MCS **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_EPICS_MCS pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mcs->record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
    "The MX_RECORD pointer for the MX_MCS pointer passed by '%s' is NULL.",
			calling_fname );
	}

	*epics_mcs = (MX_EPICS_MCS *) mcs->record->record_type_struct;

	if ( *epics_mcs == (MX_EPICS_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_EPICS_MCS pointer for record '%s' is NULL.",
			mcs->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_epics_mcs_initialize_type( long record_type )
{
	const char fname[] = "mxd_epics_mcs_initialize_type()";

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	long num_record_fields;
	long max_num_scalers_varargs_cookie;
	long max_num_measurements_varargs_cookie;
	long referenced_field_index;
	mx_status_type status;

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr
			= driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	num_record_fields = *(driver->num_record_fields);

	/* Set varargs cookies in 'data_array' that depend on the values
	 * of 'max_num_scalers' and 'max_num_measurements'.
	 */

	status = mx_find_record_field_defaults( record_field_defaults,
			num_record_fields, "data_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_find_record_field_defaults_index(
			record_field_defaults, num_record_fields,
			"max_num_scalers", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie( referenced_field_index, 0,
					&max_num_scalers_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_find_record_field_defaults_index(
			record_field_defaults, num_record_fields,
			"max_num_measurements", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie( referenced_field_index, 0,
					&max_num_measurements_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = max_num_scalers_varargs_cookie;
	field->dimension[1] = max_num_measurements_varargs_cookie;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_mcs_create_record_structures()";

	MX_MCS *mcs;
	MX_EPICS_MCS *epics_mcs;

	/* Allocate memory for the necessary structures. */

	mcs = (MX_MCS *) malloc( sizeof(MX_MCS) );

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCS structure." );
	}

	epics_mcs = (MX_EPICS_MCS *) malloc( sizeof(MX_EPICS_MCS) );

	if ( epics_mcs == (MX_EPICS_MCS *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_MCS structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mcs;
	record->record_type_struct = epics_mcs;
	record->class_specific_function_list = &mxd_epics_mcs_mcs_function_list;

	mcs->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_finish_record_initialization( MX_RECORD *record )
{
#if 0
	const char fname[] = "mxd_epics_mcs_finish_record_initialization()";
#endif

	MX_MCS *mcs;

	mx_status_type mx_status;

	mx_status = mx_mcs_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mcs = (MX_MCS *) record->record_class_struct;

	mcs->first_scaler_number = 1;

#if 0
	MX_DEBUG(-2,("%s: record = %p", fname, record));
	MX_DEBUG(-2,("%s: record->name = '%s'", fname, record->name));
	MX_DEBUG(-2,("%s: mcs = %p", fname, mcs));
	MX_DEBUG(-2,("%s: mcs->data_array = %p", fname, mcs->data_array));
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_mcs_open()";

	MX_MCS *mcs;
	MX_EPICS_MCS *epics_mcs;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	double version_number;
	int ezca_status;
	mx_status_type mx_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	mcs = (MX_MCS *) (record->record_class_struct);

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for the record's version number.
	 */

	strncpy( pvname, epics_mcs->epics_record_namestem, sizeof(pvname)-6 );
	strcat( pvname, "1.VERS" );

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &version_number );

	MX_DEBUG( 2, ("(%s) epics MCS version number = %g",
			pvname, version_number));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error reading EPICS PV '%s' for MX record '%s'.  EZCA status = %d (%s )",
			pvname, record->name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	epics_mcs->epics_record_version = version_number;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_epics_mcs_start( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_start()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long start;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( pvname, "%s1Start.VAL", epics_mcs->epics_record_namestem );

	start = 1;

	MX_DEBUG( 1,("%s: sending %ld to '%s'", fname, start, pvname));

	/* We must send this command without blocking, so we use
	 * ezcaPutOldCa().
	 */

	ezca_status = ezcaPutOldCa( pvname, ezcaLong, 1, &start );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Error starting the Struck MCS '%s' (%s).  EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG( 1,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_stop( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_stop()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long stop;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( pvname, "%s1Start.VAL", epics_mcs->epics_record_namestem );

	stop = 0;

	MX_DEBUG( 1,("%s: sending %ld to '%s'", fname, stop, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &stop );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error trying to stop the Struck MCS '%s' (%s).  EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}
	MX_DEBUG( 1,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_clear( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_clear()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long clear;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( pvname, "%s1.ERAS", epics_mcs->epics_record_namestem );

	clear = 1;

	MX_DEBUG( 1,("%S: Sending %ld to '%s'", fname, clear, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &clear );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Error clearing the Struck MCS '%s' (%s).  EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG( 1,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_busy( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_busy()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long busy;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	sprintf( pvname, "%s1Start.VAL", epics_mcs->epics_record_namestem );

	MX_DEBUG( 1,("%s: reading '%s'.", fname, pvname));

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &busy );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error while checking to see if the Struck MCS '%s' (%s) is still counting.  "
"EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname,
			mx_error_message);
	}

	MX_DEBUG( 1,("%s: busy = %ld", fname, busy));

	if ( busy ) {
		mcs->busy = TRUE;
	} else {
		mcs->busy = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_all( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_read_all()";

	long i;
	mx_status_type mx_status;

	if ( mcs == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MCS pointer passed was NULL." );
	}

	for ( i = 0; i < mcs->num_scalers; i++ ) {

		mcs->scaler_index = i;

		mx_status = mxd_epics_mcs_read_scaler( mcs );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_scaler( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_read_scaler()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long scaler_number;
	long read_cmd;
	unsigned long *data_ptr;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Please remember:  The EPICS MCS record numbers scalers starting
	 * at 1, but the mcs->data_array numbers scalers starting at 0.
	 */

	scaler_number = mcs->scaler_index + 1;

	sprintf( pvname, "%s%ld.READ",
			epics_mcs->epics_record_namestem, scaler_number );

	read_cmd = 1;

	MX_DEBUG( 1,("%s: sending %ld to '%s'", fname, read_cmd, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &read_cmd );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error reading values for Struck MCS channel %ld.  EZCA status = %d (%s )",
			scaler_number, ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	sprintf( pvname, "%s%ld.VAL",
			epics_mcs->epics_record_namestem, scaler_number );

#if 0
	MX_DEBUG(-2,("%s: mcs->record = %p", fname, mcs->record));
	MX_DEBUG(-2,("%s: mcs->record->name = '%s'", fname, mcs->record->name));
	MX_DEBUG(-2,("%s: mcs = %p", fname, mcs));
	MX_DEBUG(-2,("%s: mcs->data_array = %p", fname, mcs->data_array));
	MX_DEBUG(-2,("%s: mcs->scaler_index = %ld", fname, mcs->scaler_index));
#endif

	data_ptr = mcs->data_array[ mcs->scaler_index ];

#if 0
	MX_DEBUG(-2,("%s: data_ptr = %p", fname, data_ptr));

	MX_DEBUG(-2,("%s: mcs->num_measurements = %ld",
			fname, mcs->num_measurements));
#endif

	MX_DEBUG( 1,("%s: reading '%s'.", fname, pvname));

	ezca_status = ezcaGet( pvname, ezcaLong, 
					mcs->num_measurements, data_ptr );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error transferring data for Struck MCS channel %ld.  EZCA status = %d (%s )",
			scaler_number, ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	MX_DEBUG( 1,("%s: data received for '%s'.", fname, pvname));

#if 0
	{
		long i;

		for ( i = 0; i < mcs->num_measurements; i++ ) {
			fprintf(stderr,"%ld ", data_ptr[i]);
		}
		fprintf(stderr,"\n");
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_mcs_read_measurement( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_read_measurement()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"This function is not supported for EPICS multichannel scalers." );
}

MX_EXPORT mx_status_type
mxd_epics_mcs_get_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_get_parameter()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_epics_mcs_set_parameter( MX_MCS *mcs )
{
	const char fname[] = "mxd_epics_mcs_set_parameter()";

	MX_EPICS_MCS *epics_mcs;
	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	double dwell_time, preset_live_time;
	mx_status_type mx_status;

	mx_status = mxd_epics_mcs_get_pointers( mcs, &epics_mcs, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 1,("%s invoked for MCS '%s', type = %d",
		fname, mcs->record->name, mcs->parameter_type));

	if ( mcs->parameter_type == MXLV_MCS_MODE ) {

		if ( mcs->mode != MXM_PRESET_TIME ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal MCS mode %d selected.  Only preset time mode is "
		"allowed for an EPICS MCS.", mcs->mode );
		}

		mx_status = mxd_epics_mcs_configure( mcs, epics_mcs );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

	} else if ( mcs->parameter_type == MXLV_MCS_MEASUREMENT_TIME ) {

		sprintf( pvname, "%s1.DWEL", epics_mcs->epics_record_namestem );

		dwell_time = mcs->measurement_time;

		MX_DEBUG( 1,("%s: sending %g to '%s'",
					fname, dwell_time, pvname));

		ezca_status = ezcaPut( pvname, ezcaDouble, 1, &dwell_time );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error setting EPICS MCS '%s' (%s) dwell time.  EZCA status = %d (%s )",
				mcs->record->name,
				epics_mcs->epics_record_namestem,
				ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
					fname, mx_error_message);
		}

	} else if ( mcs->parameter_type == MXLV_MCS_NUM_MEASUREMENTS ) {

		sprintf( pvname, "%s1.PLTM", epics_mcs->epics_record_namestem );

		preset_live_time = mcs->measurement_time
					* (double) mcs->num_measurements;

		MX_DEBUG( 1,("%s: sending %g to '%s'",
					fname, preset_live_time, pvname));

		ezca_status = ezcaPut( pvname, ezcaDouble,
						1, &preset_live_time );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Error setting EPICS MCS '%s' (%s) preset live time.  EZCA status = %d (%s )",
				mcs->record->name,
				epics_mcs->epics_record_namestem,
				ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
					fname, mx_error_message);
		}

	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
		"Parameter type %d is not supported by this driver.",
			mcs->parameter_type );
	}
	MX_DEBUG( 1,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*-------------------------------------------------------------------------*/

static mx_status_type
mxd_epics_mcs_configure( MX_MCS *mcs, MX_EPICS_MCS *epics_mcs )
{
	const char fname[] = "mxd_epics_mcs_configure()";

	char pvname[80];
	char *ezca_error_message;
	char mx_error_message[150];
	int ezca_status;
	long long_value;

	MX_DEBUG( 1,("%s invoked.", fname));

	/* Put the Struck EPICS record into MCS mode. */

	sprintf( pvname, "%s1.MODE", epics_mcs->epics_record_namestem );

	long_value = 1;

	MX_DEBUG( 1,("%s: sending %ld to '%s'", fname, long_value, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error putting EPICS MCS '%s' (%s) into MCS mode.  EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Set the channel advance source for the Struck MCS. */

	sprintf( pvname, "%s1.CHAS", epics_mcs->epics_record_namestem );

	long_value = 0;

	MX_DEBUG( 1,("%s: sending %ld to '%s'", fname, long_value, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error setting the channel advance source for the EPICS MCS '%s' (%s).  "
"EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Set the prescale for the Struck MCS. */

	sprintf( pvname, "%s1.PSCL", epics_mcs->epics_record_namestem );

	long_value = 0;

	MX_DEBUG( 1,("%s: sending %ld to '%s'", fname, long_value, pvname));

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &long_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error setting the prescale for the EPICS MCS '%s' (%s).  "
"EZCA status = %d (%s )",
			mcs->record->name,
			epics_mcs->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}
	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

