/*
 * Name:    d_epscal.c
 *
 * Purpose: MX scaler driver for the EPICS APS-XFD developed scaler record.
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_meas.h"
#include "mx_scalr.h"
#include "d_epscal.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the scaler driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_scaler_record_function_list = {
	mxd_epics_scaler_initialize_type,
	mxd_epics_scaler_create_record_structures,
	mxd_epics_scaler_finish_record_initialization,
	mxd_epics_scaler_delete_record,
	mxd_epics_scaler_print_structure,
	mxd_epics_scaler_read_parms_from_hardware,
	mxd_epics_scaler_write_parms_to_hardware,
	mxd_epics_scaler_open,
	mxd_epics_scaler_close
};

MX_SCALER_FUNCTION_LIST mxd_epics_scaler_scaler_function_list = {
	mxd_epics_scaler_clear,
	mxd_epics_scaler_overflow_set,
	mxd_epics_scaler_read,
	mxd_epics_scaler_read_and_clear,
	mxd_epics_scaler_is_busy,
	mxd_epics_scaler_start,
	mxd_epics_scaler_stop,
	mxd_epics_scaler_get_mode,
	mxd_epics_scaler_set_mode,
	mxd_epics_scaler_set_modes_of_associated_counters
};

/* EPICS scaler data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_scaler_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_SCALER_STANDARD_FIELDS,
	MXD_EPICS_SCALER_STANDARD_FIELDS
};

long mxd_epics_scaler_num_record_fields
		= sizeof( mxd_epics_scaler_record_field_defaults )
		  / sizeof( mxd_epics_scaler_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_scaler_rfield_def_ptr
			= &mxd_epics_scaler_record_field_defaults[0];

MX_EXPORT mx_status_type
mxd_epics_scaler_initialize_type( long type )
{
	/* Nothing here yet. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_scaler_create_record_structures()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;

	/* Allocate memory for the necessary structures. */

	scaler = (MX_SCALER *) malloc( sizeof(MX_SCALER) );

	if ( scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCALER structure." );
	}

	epics_scaler = (MX_EPICS_SCALER *)
				malloc( sizeof(MX_EPICS_SCALER) );

	if ( epics_scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_SCALER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = scaler;
	record->record_type_struct = epics_scaler;
	record->class_specific_function_list
			= &mxd_epics_scaler_scaler_function_list;

	scaler->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_finish_record_initialization( MX_RECORD *record )
{
	const char fname[]
		= "mxd_epics_scaler_finish_record_initialization()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;
	int scaler_number;

	scaler = (MX_SCALER *)( record->record_class_struct );

	epics_scaler = (MX_EPICS_SCALER *)( record->record_type_struct );

	/* Check that the scaler number is valid. */

	scaler_number = epics_scaler->scaler_number;

	if ( scaler_number < 1 || scaler_number > 16 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"EPICS scaler channel number %d is out of allowed range 2-16",
			scaler_number );
	}

	if ( scaler_number == 1 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"EPICS scaler channel 1 is reserved for timer use "
		"and cannot be used as a normal scaler.  Use channels 2-16" );
	}

	/* EPICS scalers don't seem to have a way of testing whether
	 * their value has overflowed.
	 */

	scaler->automatic_overflow_check = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_epics_scaler_print_structure()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;
	long current_value;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	scaler = (MX_SCALER *) (record->record_class_struct);

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SCALER pointer for record '%s' is NULL.", record->name);
	}

	epics_scaler = (MX_EPICS_SCALER *) (record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_EPICS_SCALER pointer for record '%s' is NULL.", record->name);
	}

	fprintf(file, "SCALER parameters for scaler '%s':\n", record->name);

	fprintf(file, "  Scaler type           = EPICS_SCALER.\n\n");
	fprintf(file, "  EPICS record          = %s\n",
					epics_scaler->epics_record_name);
	fprintf(file, "  scaler number         = %d\n",
					epics_scaler->scaler_number);

	status = mx_scaler_read( record, &current_value );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read value of scaler '%s'",
			record->name );
	}

	fprintf(file, "  present value         = %ld\n", current_value);
	fprintf(file, "  dark current          = %g counts per second.\n",
						scaler->dark_current);
	fprintf(file, "  subtract_dark_current = %d\n",
						scaler->subtract_dark_current);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_scaler_open()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	double version_number;
	int ezca_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	scaler = (MX_SCALER *) (record->record_class_struct);

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SCALER pointer for record '%s' is NULL.",
			record->name);
	}

	epics_scaler = (MX_EPICS_SCALER *) (record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_EPICS_SCALER pointer for scaler record '%s' is NULL.",
			record->name );
	}

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for the record's version number.
	 */

	strncpy( pvname, epics_scaler->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".VERS" );

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &version_number );

	MX_DEBUG( 2, ("(%s) epics scaler version number = %g",
		epics_scaler->epics_record_name, version_number));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
		    "Error getting EPICS device type. EZCA status = %d (%s )",
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	epics_scaler->epics_record_version = version_number;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_clear( MX_SCALER *scaler )
{
	/* It appears that it is not possible to directly clear
	 * an EPICS scaler channel.  So all we can do is return
	 * a success code.
	 */

	scaler->value = 0L;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_overflow_set( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_overflow_set()";

	scaler->overflow_set = 0;

	return mx_error( MXE_UNSUPPORTED, fname,
		"This type of scaler cannot check for overflow set." );
}

MX_EXPORT mx_status_type
mxd_epics_scaler_read( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_read()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long value;
	int ezca_status;

	epics_scaler = (MX_EPICS_SCALER *)(scaler->record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"EPICS_SCALER pointer is NULL.");
	}

	sprintf( pvname, "%s.S%d", epics_scaler->epics_record_name,
				epics_scaler->scaler_number );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read from EPICS scaler '%s' (%s) channel %d failed. "
"EZCA status = %d (%s )",
			scaler->record->name,
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		scaler->value = 0L;

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	scaler->value = value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_read_and_clear( MX_SCALER *scaler )
{
	mx_status_type status;

	status = mxd_epics_scaler_read( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxd_epics_scaler_clear( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_is_busy( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_is_busy()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long count_field;
	int ezca_status;

	epics_scaler = (MX_EPICS_SCALER *) (scaler->record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_SCALER structure pointer is NULL.");
	}

	sprintf( pvname, "%s.CNT", epics_scaler->epics_record_name );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read busy status of EPICS scaler '%s' (%s, channel %d) failed. "
"EZCA status = %d (%s )",
			scaler->record->name,
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	switch ( count_field ) {
	case 0:
		scaler->busy = FALSE;
		break;
	case 1:
		scaler->busy = TRUE;
		break;
	default:
		scaler->busy = FALSE;

		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Count field (.CNT) for scaler '%s' had an unexpected value of %d",
			epics_scaler->epics_record_name, count_field );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_start( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_start()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long value, count_field;
	int ezca_status;

	epics_scaler = (MX_EPICS_SCALER *)(scaler->record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"EPICS_SCALER pointer is NULL.");
	}

	/* Set the scaler preset. */

	value = scaler->value;

	sprintf( pvname, "%s.PR%d", epics_scaler->epics_record_name,
				epics_scaler->scaler_number );

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to write to EPICS scaler '%s' (%s, channel %d) failed.  "
"EZCA status = %d (%s )",
			scaler->record->name,
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Start the scaler counting. */

	sprintf( pvname, "%s.CNT", epics_scaler->epics_record_name );

	count_field = 1;

	if ( epics_scaler->epics_record_version ) {
		ezca_status = ezcaPut( pvname, ezcaLong, 1, &count_field );
	} else {
		ezca_status = ezcaPut( pvname, ezcaLong, 1, &count_field );
	}

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to start EPICS scaler '%s' (%s, channel %d) failed.  "
"EZCA status = %d (%s )",
			scaler->record->name,
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_stop( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_stop()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long count_field;

	epics_scaler = (MX_EPICS_SCALER *)
				( scaler->record->record_type_struct );

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_SCALER structure pointer is NULL.");
	}

	/* Stop the scaler from counting. */

	sprintf( pvname, "%s.CNT", epics_scaler->epics_record_name );

	count_field = 0;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to stop EPICS scaler '%s' (%s, channel %d) failed.  "
"EZCA status = %d (%s )",
			scaler->record->name,
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	scaler->value = 0.0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_get_mode( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_get_mode()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long gate_control;

	epics_scaler = (MX_EPICS_SCALER *)
				( scaler->record->record_type_struct );

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_SCALER structure pointer is NULL.");
	}

	/* Get the gate control field G1. */

	sprintf( pvname, "%s.G%d", epics_scaler->epics_record_name,
					epics_scaler->scaler_number );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &gate_control );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
			"Attempt to get the gate control field for "
			"EPICS counter '%s' (%s, channel %d) failed.  "
			"EZCA status = %d (%s )",

			scaler->record->name,
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	switch( gate_control ) {
	case 0:
		scaler->mode = MXCM_COUNTER_MODE;
		break;

	case 1:
		scaler->mode = MXCM_PRESET_MODE;
		break;

	default:
		return mx_error( MXE_CONTROLLER_INTERNAL_ERROR, fname,
		"The gate control field for EPICS scaler '%s' (%s) "
		"has the unexpected value %ld.",
			scaler->record->name,
			epics_scaler->epics_record_name,
			gate_control );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_set_mode( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_set_mode()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long gate_control, counter_mode;
	mx_status_type mx_status;

	epics_scaler = (MX_EPICS_SCALER *)
				( scaler->record->record_type_struct );

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_SCALER structure pointer is NULL.");
	}

	/* Choose the correct gate control mode. */

	switch( scaler->mode ) {
	case MXCM_PRESET_MODE:
		gate_control = 1;
		break;

	case MXCM_COUNTER_MODE:
		gate_control = 0;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The counter mode %ld requested for scaler '%s' is not supported.",
			scaler->mode, scaler->record->name );
		break;
	}

	/* Make sure the counter is _not_ in continuous count mode. */

	if ( epics_scaler->epics_record_version < 3.0 ) {
		sprintf( pvname,
			"%s_mode.VAL", epics_scaler->epics_record_name );

		counter_mode = 0;	/* One-shot mode. */

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &counter_mode );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
				"Attempt to set one-shot mode for "
				"EPICS scaler '%s' (%s, channel %d) failed.  "
				"EZCA status = %d (%s )",

				scaler->record->name,
				epics_scaler->epics_record_name,
				epics_scaler->scaler_number,
				ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
						fname, mx_error_message);
		}
	}

	/* Tell the scaler to stop counting. */

	mx_status = mxd_epics_scaler_stop( scaler );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Set the gate control field. */

	sprintf( pvname, "%s.G%d", epics_scaler->epics_record_name,
					epics_scaler->scaler_number );

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &gate_control );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
			"Attempt to set the gate control field for "
			"EPICS counter '%s' (%s, channel %d) failed.  "
			"EZCA status = %d (%s )",

			scaler->record->name,
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_set_modes_of_associated_counters( MX_SCALER *scaler )
{
	const char fname[]
		= "mxd_epics_scaler_set_modes_of_associated_counters()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int i, ezca_status;
	long gate_control;

	epics_scaler = (MX_EPICS_SCALER *)
				scaler->record->record_type_struct;

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_SCALER structure pointer is NULL.");
	}

	switch( scaler->mode ) {
	case MXCM_PRESET_MODE:
	case MXCM_COUNTER_MODE:
		break;
	default:
		return mx_error( MXE_UNSUPPORTED, fname,
		"The counter mode %d is not supported by this function.",
			scaler->mode );
		break;
	}

	/* Start an EPICS synchronous group. */

	ezca_status = ezcaStartGroup();

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
		"Attempt to start an EPICS synchronous group for record '%s' "
		"failed.  EZCA status = %d (%s )",
			scaler->record->name,
			ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR,
			fname, mx_error_message);
	}

	/* Set the gate control fields for all channels but this one. */

	for ( i = 1; i <= 16; i++ ) {

		/* Skip the channel that corresponds to this scaler. */

		if ( i == epics_scaler->scaler_number ) {
			continue;
		}

		if ( i > 1 ) {
			/* The other scaler channels are always put into
			 * counter mode.
			 */

			gate_control = 0;

		} else {
			/* Channel 1 is the timer.  The timer's gate mode
			 * is always set to the opposite of the counter mode
			 * for the scaler_record variable.
			 */

			if ( scaler->mode == MXCM_PRESET_MODE ) {
				gate_control = 0;	/* counter mode */
			} else {
				gate_control = 1;	/* preset mode */
			}
		}

		sprintf( pvname, "%s.G%d", epics_scaler->epics_record_name, i );

		ezca_status = ezcaPut( pvname, ezcaLong, 1, &gate_control );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
			    "Attempt to set the gate control field for the "
			    "associated channel %d of EPICS counter '%s' (%s) "
			    "failed.  EZCA status = %d (%s )",
			    i, scaler->record->name,
			    epics_scaler->epics_record_name,
			    ezca_status, ezca_error_message );
			    ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
				fname, mx_error_message);
		}
	}

	/* Send the synchronous group to EPICS. */

	ezca_status = ezcaEndGroup();

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
		"Execution of EPICS synchronous group for record '%s' "
		"failed.  EZCA status = %d (%s )",
			scaler->record->name,
			ezca_status, ezca_error_message );
			ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR,
			fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

