/*
 * Name:   i_pmac.h
 *
 * Purpose: Header for MX driver for Delta Tau PMAC controllers.  Note that
 *          this is the driver for talking to the PMAC itself.  PMAC-controlled
 *          motors are interfaced via the driver in libmx/d_pmac.c.
 *
 * Author:  William Lavender
 *
 *----------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef __I_PMAC_H__
#define __I_PMAC_H__

#include "mx_rec.h"
#include "mx_motor.h"
#include "mx_gener.h"
#include "mx_rs232.h"

/* Define the data structures used by a PMAC controller. */

#define MX_PMAC_TYPE_PMAC1		1
#define MX_PMAC_TYPE_PMAC2		2
#define MX_PMAC_TYPE_PMACUL		3

#define MX_PMAC_TYPE_TURBO		1024

#define MX_PMAC_TYPE_TURBO1		( 1 | MX_PMAC_TYPE_TURBO )
#define MX_PMAC_TYPE_TURBO2		( 2 | MX_PMAC_TYPE_TURBO )
#define MX_PMAC_TYPE_TURBOU		( 3 | MX_PMAC_TYPE_TURBO )

/*---*/

#define MX_PMAC_PORT_TYPE_RS232		1
#define MX_PMAC_PORT_TYPE_EPICS_TC	11

#define MX_PMAC_PORT_TYPE_LENGTH	32
#define MX_PMAC_PORT_ARGS_LENGTH	80

typedef struct {
	MX_RECORD *record;
	int pmac_type;
	int port_type;

	/* PMAC controller PROM version. */

	int major_version;
	int minor_version;

	int i6_variable;  /* affects error reporting */

	/* Parameters shared by all eight motor axes. */

	MX_RECORD *port_record;

	char port_type_name[MX_PMAC_PORT_TYPE_LENGTH+1];
	char port_args[MX_PMAC_PORT_ARGS_LENGTH+1];
	int num_cards;

} MX_PMAC;

#define MXI_PMAC_STANDARD_FIELDS \
  {-1, -1, "port_type_name", MXFT_STRING, NULL, 1,{MX_PMAC_PORT_TYPE_LENGTH},\
	MXF_REC_TYPE_STRUCT, offsetof(MX_PMAC, port_type_name), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}, \
  \
  {-1, -1, "port_args", MXFT_STRING, NULL, 1, {MX_PMAC_PORT_ARGS_LENGTH},\
	MXF_REC_TYPE_STRUCT, offsetof(MX_PMAC, port_args), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}, \
  \
  {-1, -1, "num_cards", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_TYPE_STRUCT, offsetof(MX_PMAC, num_cards), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}

#ifndef MX_CR
#define MX_ACK		'\006'
#define MX_BELL		'\007'
#define MX_LF		'\012'
#define MX_CR		'\015'
#define MX_CTRL_Z	'\032'
#endif

MX_API mx_status_type mxi_pmac_initialize_type( long type );
MX_API mx_status_type mxi_pmac_create_record_structures( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_finish_record_initialization(
							MX_RECORD *record );
MX_API mx_status_type mxi_pmac_delete_record( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_read_parms_from_hardware( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_write_parms_to_hardware( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_open( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_close( MX_RECORD *record );
MX_API mx_status_type mxi_pmac_resynchronize( MX_RECORD *record );

MX_API mx_status_type mxi_pmac_getchar( MX_GENERIC *generic,
					char *c, int flags );
MX_API mx_status_type mxi_pmac_putchar( MX_GENERIC *generic,
					char c, int flags );
MX_API mx_status_type mxi_pmac_read( MX_GENERIC *generic,
					void *buffer, size_t count );
MX_API mx_status_type mxi_pmac_write( MX_GENERIC *generic,
					void *buffer, size_t count );
MX_API mx_status_type mxi_pmac_input_is_available( MX_GENERIC *generic,
						int *input_available );
MX_API mx_status_type mxi_pmac_discard_unread_input( MX_GENERIC *generic,
							int debug_flag );
MX_API mx_status_type mxi_pmac_discard_unwritten_output( MX_GENERIC *generic,
							int debug_flag );

extern MX_RECORD_FUNCTION_LIST mxi_pmac_record_function_list;
extern MX_GENERIC_FUNCTION_LIST mxi_pmac_generic_function_list;

extern long mxi_pmac_num_record_fields;
extern MX_RECORD_FIELD_DEFAULTS *mxi_pmac_rfield_def_ptr;

/* === Driver specific functions === */

MX_API mx_status_type mxi_pmac_command( MX_PMAC *pmac, char *command,
		char *response, int response_buffer_length, int debug_flag );

MX_API mx_status_type mxi_pmac_send_command( MX_PMAC *pmac,
					char *command, int debug_flag );

MX_API mx_status_type mxi_pmac_receive_response( MX_PMAC *pmac, 
		char *response, int response_buffer_length, int debug_flag );

#endif /* __I_PMAC_H__ */
