/*
 * Name:    i_tty.c
 *
 * Purpose: MX driver for Unix TTY RS-232 devices.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mx_osdef.h"

#ifdef OS_UNIX	/* Don't include this on non-Unix systems. */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>

#include "mx_util.h"
#include "mx_rs232.h"
#include "mx_rec.h"
#include "i_tty.h"

#include "select.h"

#if defined(OS_LINUX) || defined(OS_SOLARIS2) || defined(OS_SGI) \
	|| defined(OS_SUNOS4) || defined(OS_AIX) || defined(OS_HPUX)
#    define USE_POSIX_TERMIOS	TRUE
#  else
#    error "No Unix TTY handling interface has been defined."
#endif

/* Do we show chars using MX_DEBUG() in the getchar and putchar functions? */

#define SHOW_CHARS			FALSE

#if USE_POSIX_TERMIOS
#include <termios.h>

/* Define private functions to handle POSIX termios style terminal I/O. */

#if 0
static mx_status_type posix_termios_read_parms( MX_RS232 *rs232 );
#endif

static mx_status_type posix_termios_write_parms( MX_RS232 *rs232 );

#endif /* USE_POSIX_TERMIOS */

MX_RECORD_FUNCTION_LIST mxi_tty_record_function_list = {
	mxi_tty_initialize_type,
	mxi_tty_create_record_structures,
	mxi_tty_finish_record_initialization,
	mxi_tty_delete_record,
	NULL,
	mxi_tty_read_parms_from_hardware,
	mxi_tty_write_parms_to_hardware,
	mxi_tty_open,
	mxi_tty_close
};

MX_RS232_FUNCTION_LIST mxi_tty_rs232_function_list = {
	mxi_tty_getchar,
	mxi_tty_putchar,
	mxi_tty_read,
	mxi_tty_write,
	NULL,
	NULL,
	mxi_tty_input_is_available,
	mxi_tty_discard_unread_input,
	mxi_tty_discard_unwritten_output
};

MX_RECORD_FIELD_DEFAULTS mxi_tty_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_RS232_STANDARD_FIELDS,
	MXI_TTY_STANDARD_FIELDS
};

long mxi_tty_num_record_fields
		= sizeof( mxi_tty_record_field_defaults )
			/ sizeof( mxi_tty_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_tty_rfield_def_ptr
			= &mxi_tty_record_field_defaults[0];

/* ---- */

MX_EXPORT mx_status_type
mxi_tty_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_tty_create_record_structures()";

	MX_RS232 *rs232;
	MX_TTY *tty;

	/* Allocate memory for the necessary structures. */

	rs232 = (MX_RS232 *) malloc( sizeof(MX_RS232) );

	if ( rs232 == (MX_RS232 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_RS232 structure." );
	}

	tty = (MX_TTY *) malloc( sizeof(MX_TTY) );

	if ( tty == (MX_TTY *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_TTY structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = rs232;
	record->record_type_struct = tty;
	record->class_specific_function_list
				= &mxi_tty_rs232_function_list;

	rs232->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_tty_finish_record_initialization()";

	MX_TTY *tty;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	tty = (MX_TTY *) (record->record_type_struct);

	if ( tty == (MX_TTY *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TTY structure pointer for '%s' is NULL.", record->name );
	}

	/* Check to see if the RS-232 parameters are valid. */

	status = mx_rs232_check_port_parameters( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Mark the TTY device as being closed. */

	tty->file_handle = -1;

	record->flags |= ( ~MXF_REC_OPEN );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_read_parms_from_hardware( MX_RECORD *record )
{
	MX_DEBUG( 2,
	("mxi_tty_read_parms_from_hardware() not yet implemented."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_write_parms_to_hardware( MX_RECORD *record )
{
	const char fname[] = "mxi_tty_write_parms_to_hardware()";

	MX_RS232 *rs232;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	rs232 = (MX_RS232 *) (record->record_class_struct);

	if ( rs232 == (MX_RS232 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_RS232 pointer for record '%s' is NULL.", record->name);
	}

#if USE_POSIX_TERMIOS
	status = posix_termios_write_parms( rs232 );

#else  /* USE_POSIX_TERMIOS */

#error "No TTY parameter setting mechanism is defined."

#endif /* USE_POSIX_TERMIOS */

	MX_DEBUG(2,
	("Leaving mxi_tty_write_parms_to_hardware() with status.code = %ld",
		status.code));

	return status;
}

MX_EXPORT mx_status_type
mxi_tty_open( MX_RECORD *record )
{
	const char fname[] = "mxi_tty_open()";

	MX_RS232 *rs232;
	MX_TTY *tty;
	char *name;
	int status, flags, saved_errno;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	name = &(record->name[0]);	 /* Get the name of the port. */

	rs232 = (MX_RS232 *) (record->record_class_struct);

	if ( rs232 == (MX_RS232 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_RS232 structure pointer is NULL." );
	}

	tty = (MX_TTY *) (record->record_type_struct);

	if ( tty == (MX_TTY *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_TTY structure pointer for '%s' is NULL.", name );
	}

	MX_DEBUG( 2,
("mxi_tty_open(): Somewhere in here we need to worry about TTY lock files."));

	/* If the tty device is currently open, close it. */

	if ( tty->file_handle >= 0 ) {
		status = close( tty->file_handle );

		if ( status != 0 ) {
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Error while closing previously open TTY device '%s'.", name );
		}
	}

	/* Decide on the flags for the open(). */

#if 1
	flags = O_RDWR | O_NOCTTY;
#else
	flags = O_RDWR | O_NOCTTY | O_NDELAY;
#endif

	/* Now open the tty device. */

	tty->file_handle = open( tty->filename, flags );

	if ( tty->file_handle < 0 ) {
		saved_errno = errno;

		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"Error opening TTY device '%s'.  Errno = %d, errno text = '%s'",
			name, saved_errno, strerror( saved_errno ) );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_close( MX_RECORD *record )
{
	const char fname[] = "mxi_tty_close()";

	MX_RS232 *rs232;
	MX_TTY *tty;
	char *name;
	int result;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	name = &(record->name[0]);	 /* Get the name of the port. */

	rs232 = (MX_RS232 *) (record->record_class_struct);

	if ( rs232 == (MX_RS232 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_RS232 structure pointer is NULL." );
	}

	tty = (MX_TTY *) (record->record_type_struct);

	if ( tty == (MX_TTY *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_TTY structure pointer is NULL." );
	}

	MX_DEBUG( 2,
("mxi_tty_close(): Somewhere in here we need to worry about TTY lock files."));

	/* If the tty device is currently open, close it. */

	if ( tty->file_handle >= 0 ) {
		result = close( tty->file_handle );

		if ( result != 0 ) {
			return mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Error while closing TTY device '%s'.", name);
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_getchar( MX_RS232 *rs232, char *c, int flags )
{
	const char fname[] = "mxi_tty_getchar()";

	MX_TTY *tty;
	int num_chars, saved_flags, new_flags, result, saved_errno;
	char c_temp;
	unsigned char c_mask;

#if SHOW_CHARS
	MX_DEBUG(-2, ("mxi_tty_getchar() invoked."));
#endif

	tty = (MX_TTY*) (rs232->record->record_type_struct);

	if ( flags & MXF_232_NOWAIT ) {
		/* Set the file descriptor to be non-blocking. */

		saved_flags = fcntl( tty->file_handle, F_GETFL );

		new_flags = saved_flags & O_NONBLOCK;

		if ( new_flags != saved_flags ) {
			result = fcntl( tty->file_handle, F_SETFL,
					(long) new_flags );

			if ( result == -1 ) {
				saved_errno = errno;

				return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"fcntl( ,F_SETFL, ) error.  Errno = %d, errno text = '%s'",
					saved_errno, strerror( saved_errno ) );
			}

			num_chars = read( tty->file_handle, &c_temp, 1 );

			result = fcntl( tty->file_handle, F_SETFL,
					(long) saved_flags );

			if ( result == -1 ) {
				saved_errno = errno;

				return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"fcntl( ,F_GETFL, ) error.  Errno = %d, errno text = '%s'",
					saved_errno, strerror( saved_errno ) );
			}
		} else {
			num_chars = read( tty->file_handle, &c_temp, 1 );
		}
	} else {

		num_chars = read( tty->file_handle, &c_temp, 1 );
	}

	/* === Mask the character to the appropriate number of bits. === */

	/* Note that the following code _assumes_ that chars are 8 bits. */

	c_mask = 0xff >> ( 8 - rs232->word_size );

	c_temp &= c_mask;

	*c = (char) c_temp;

#if SHOW_CHARS
	MX_DEBUG(-2, ("mxi_tty_getchar(): num_chars = %d, c = 0x%x, '%c'",
		num_chars, *c, *c));
	MX_DEBUG(-2, ("mxi_tty_getchar(2): num_chars = %d, c = 0x%x, '%c'",
		num_chars, c_temp, c_temp));
	MX_DEBUG(-2, ("mxi_tty_getchar(2): c_mask = 0x%x", c_mask));
#endif

	/* mxi_tty_getchar() is often used to test whether there is
	 * input ready on the TTY port.  Normally, it is not desirable
	 * to broadcast a message to the world when this fails, so we
	 * use mx_error_quiet() rather than mx_error().
	 */

	if ( num_chars != 1 ) {
		return mx_error_quiet( MXE_NOT_READY, fname,
			"Failed to read a character from port '%s'.",
			rs232->record->name );
	} else {
		return MX_SUCCESSFUL_RESULT;
	}
}

MX_EXPORT mx_status_type
mxi_tty_putchar( MX_RS232 *rs232, char c, int flags )
{
	const char fname[] = "mxi_tty_putchar()";

	MX_TTY *tty;
	int num_chars;

#if SHOW_CHARS
	MX_DEBUG(-2, ("mxi_tty_putchar() invoked.  c = 0x%x, '%c'", c, c));
#endif

	tty = (MX_TTY*) (rs232->record->record_type_struct);

	if ( flags & MXF_232_NOWAIT ) {
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"Non-blocking TTY I/O not yet implemented.");
	} else {

		num_chars = write( tty->file_handle, &c, 1 );
	}

#if SHOW_CHARS
	MX_DEBUG(-2, ("mxi_tty_putchar(): num_chars = %d", num_chars));
#endif

	if ( num_chars != 1 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"Attempt to send a character to port '%s' failed.  character = '%c'",
			rs232->record->name, c );
	} else {
		return MX_SUCCESSFUL_RESULT;
	}
}

MX_EXPORT mx_status_type
mxi_tty_read( MX_RS232 *rs232,
		char *buffer,
		size_t max_bytes_to_read,
		size_t *bytes_read )
{
	mx_status_type status;
	long i;

	/* This is slow, but simple.  It currently waits forever until
	 * 'max_chars' are read.  Think about doing something better later.
	 */

	for ( i = 0; i < max_bytes_to_read; i++ ) {
		status = mxi_tty_getchar( rs232, &buffer[i], MXF_232_WAIT );

		if ( status.code != MXE_SUCCESS ) {
			i--;
			continue;
		}
	}

	*bytes_read = max_bytes_to_read;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_write( MX_RS232 *rs232,
		char *buffer,
		size_t max_bytes_to_write,
		size_t *bytes_written )
{
	mx_status_type status;
	long i;

	/* This is slow, but simple.  It currently waits forever until
	 * 'max_chars' are written.  Think about doing something 
	 * better later.
	 */

	for ( i = 0; i < max_bytes_to_write; i++ ) {
		status = mxi_tty_putchar( rs232, buffer[i], MXF_232_WAIT );

		if ( status.code != MXE_SUCCESS ) {
			i--;
			continue;
		}
	}

	*bytes_written = max_bytes_to_write;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_input_is_available( MX_RS232 *rs232, int *input_available )
{
	const char fname[] = "mxi_tty_input_is_available()";

	MX_TTY *tty;
	int tty_fd, select_status;
	struct timeval timeout;

#if USE_FD_SET
	fd_set mask;

	FD_ZERO( &mask );

#else /* not USE_FD_SET */

	long mask;

#endif /* USE_FD_SET */

	MX_DEBUG( 2, ("%s invoked.", fname));

	tty = (MX_TTY *) (rs232->record->record_type_struct);

	if ( tty == (MX_TTY *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TTY structure for TTY port '%s' is NULL.",
			rs232->record->name);
	}

	tty_fd = tty->file_handle;

	if ( tty_fd < 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"TTY port device '%s' is not open.  TTY fd = %d",
		rs232->record->name, tty_fd);
	}

#if USE_FD_SET

	FD_SET( tty_fd, &mask );
#else
	mask = 1 << tty_fd;
#endif

	/* Set timeout to zero. */

	timeout.tv_usec = 0;
	timeout.tv_sec = 0;
	
	/* Do the test. */

	select_status = select( 1 + tty_fd, &mask, NULL, NULL, &timeout );

	if ( select_status ) {
		*input_available = TRUE;
	} else {
		*input_available = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_discard_unread_input( MX_RS232 *rs232, int debug_flag )
{
	const char fname[] = "mxi_tty_discard_unread_input()";

	int input_available;
	unsigned long i, timeout;
	char c;
	mx_status_type status;

	char buffer[100];
	unsigned long j, k;
	unsigned long buffer_length = sizeof(buffer) / sizeof(buffer[0]);

	MX_DEBUG( 2,("%s invoked.", fname));

	timeout = 10000L;

	/* If input is available, read until there is no more input.
	 * If we do this to a device that is constantly generating
	 * output, then we will never get to the end.  Thus, we have
	 * built a timeout into the function.
	 */

	for ( i=0; i < timeout; i++ ) {

		mx_msleep(5);

		status = mxi_tty_input_is_available( rs232, &input_available );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( input_available == FALSE ) {

			break;	/* Here we exit the for() loop. */

		} else {
			status = mxi_tty_getchar( rs232, &c, MXF_232_WAIT );

			if ( status.code != MXE_SUCCESS )
				return status;

			if ( debug_flag ) {
				if ( i < buffer_length ) {
					buffer[i] = c;
				}
			}
		}
	}

	if ( debug_flag ) {
		if ( i > buffer_length )
			k = buffer_length;
		else
			k = i;

		if ( i > 0 ) {
			fprintf(stderr,
				"%s: First %lu characters discarded were:\n\n",
				fname,k);

			for ( j = 0; j < k; j++ ) {
				c = buffer[j];

				if ( isprint( (int) c ) ) {
					fputc( c, stderr );
				} else {
					fprintf( stderr, "(0x%x)", c );
				}
			}

			fprintf(stderr,"\n\n");
		}
	}

	if ( i >= timeout ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"As many as %lu characters were read while trying to empty the input buffer.  "
"Perhaps this device continuously generates output?", timeout );

	} else if ( debug_flag && (i > 0) ) {

		MX_DEBUG(-2,("%s: %lu characters discarded.", fname, i));
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_tty_discard_unwritten_output( MX_RS232 *rs232, int debug_flag )
{
	const char fname[] = "mxi_tty_discard_unwritten_output()";

	MX_DEBUG( 2,("%s invoked.", fname));

	/* This is not generally possible for a Unix TTY device. */

	return mx_error_quiet( MXE_UNSUPPORTED, fname,
		"This function is not supported for a Unix TTY device." );
}

/* ==== Unix version dependent I/O functions ==== */

#if USE_POSIX_TERMIOS

static mx_status_type
posix_termios_write_parms( MX_RS232 *rs232 )
{
	const char fname[] = "posix_termios_write_parms()";

	MX_RECORD *record;
	MX_TTY *tty;
	int tty_fd, ispeed, ospeed;
	char *name;
	struct termios tio, tio_copy;
	int i, error;

	MX_DEBUG( 2,("%s invoked.", fname));

	/* Get the name of the port. */

	record = rs232->record;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD ptr for MX_RS232 ptr = 0x%p is NULL.", rs232);
	}

	name = &(record->name[0]);

	tty = (MX_TTY *) (record->record_type_struct);

	if ( tty == (MX_TTY *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TTY structure for TTY port '%s' is NULL.", name);
	}

	tty_fd = tty->file_handle;

	if ( tty_fd < 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"TTY port device '%s' is not open.  TTY fd = %d",
		name, tty_fd);
	}

	/* Wait for currently pending output to finish. */

	if ( tcdrain( tty_fd ) != 0 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Output to TTY '%s' never finished just before tcgetattr().", name);
	}

	/* Set the port speed. */

	switch( rs232->speed ) {
	case 38400: ispeed = ospeed = B38400; break;
	case 19200: ispeed = ospeed = B19200; break;
	case 9600:  ispeed = ospeed = B9600;  break;
	case 4800:  ispeed = ospeed = B4800;  break;
	case 2400:  ispeed = ospeed = B2400;  break;
	case 1800:  ispeed = ospeed = B1800;  break;
	case 1200:  ispeed = ospeed = B1200;  break;
	case 600:   ispeed = ospeed = B600;   break;
	case 300:   ispeed = ospeed = B300;   break;
	case 200:   ispeed = ospeed = B200;   break;
	case 150:   ispeed = ospeed = B150;   break;
	case 110:   ispeed = ospeed = B110;   break;
	case 75:    ispeed = ospeed = B75;    break;
	case 50:    ispeed = ospeed = B50;    break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported TTY speed = %d for TTY '%s'.", 
		rs232->speed, name);

		break;
	}

	/* Since the serial port speed is not set using tcsetattr(),
	 * we set the speed before doing a tcgetattr(), since setting
	 * the speed later can confuse our test for whether the
	 * termios bits were set correctly.
	 */

	cfsetispeed( &tio, ispeed );
	cfsetospeed( &tio, ospeed );

	/* Get the current status of the port so that we only 
	 * need change those parameters we need.
	 */

	if ( tcgetattr( tty_fd, &tio ) != 0 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Unable to get current state of TTY '%s' before changing its settings.",
			name);
	}

	/* = Set all the parameters that are the same for all TTY ports. = */

	tio.c_lflag &= (~ICANON);	/* No input processing. */
	tio.c_lflag &= (~ECHO);		/* No echo. */

	tio.c_iflag |= (IGNBRK);	/* Don't want break conditions. */

	tio.c_iflag &= (~INLCR);	/* Don't change NLs or CRs around. */
	tio.c_iflag &= (~IGNCR);
	tio.c_iflag &= (~ICRNL);

	tio.c_iflag &= (~ISTRIP);	/* Don't strip characters to 7 bits. */

	tio.c_oflag &= (~OPOST);	/* No output processing. */

	tio.c_cflag |= (CREAD);		/* Enable reading of characters. */

	tio.c_cc[VMIN] = 0;   /* Read one character at a time. */
	tio.c_cc[VTIME] = 10; /* Wait at most 1 second for that character. */

	/* Set the word size. */

	tio.c_cflag &= ~(CSIZE);

	switch( rs232->word_size ) {
	case 8: tio.c_cflag |= CS8; break;
	case 7: tio.c_cflag |= CS7; break;
	case 6: tio.c_cflag |= CS6; break;
	case 5: tio.c_cflag |= CS5; break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported TTY word size = %d for TTY '%s'.", 
		rs232->word_size, name);

		break;
	}

	/* Set the parity. */

	switch( rs232->parity ) {
	case MXF_232_NO_PARITY:
		tio.c_iflag &= (~INPCK);
		tio.c_cflag &= (~PARENB);
		break;

	case MXF_232_EVEN_PARITY:
		tio.c_iflag |= INPCK;
		tio.c_cflag |= PARENB;
		tio.c_cflag &= (~PARODD);
		break;
		
	case MXF_232_ODD_PARITY:
		tio.c_iflag |= INPCK;
		tio.c_cflag |= PARENB;
		tio.c_cflag |= PARODD;
		break;

	case MXF_232_MARK_PARITY:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"TTY = '%s'.  Mark parity is not supported for POSIX termios interfaces.",
			name);
		break;

	case MXF_232_SPACE_PARITY:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"TTY = '%s'.  Space parity is not supported for POSIX termios interfaces.",
			name);
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported TTY parity = '%c' for TTY '%s'.", 
		rs232->parity, name);

		break;
	}

	/* Set the stop bits. */

	switch( rs232->stop_bits ) {
	case 1: tio.c_cflag &= (~CSTOPB); break;
	case 2: tio.c_cflag |= CSTOPB;    break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported TTY stop bits = %d for TTY '%s'.", 
		rs232->stop_bits, name);

		break;
	}

	/* Set the flow control. */

	switch( rs232->flow_control ) {
	case MXF_232_NO_FLOW_CONTROL:
		/* No to hardware flow control. */
		tio.c_cflag |= CLOCAL;

		/* No to software flow control. */
		tio.c_iflag &= (~IXON);
		tio.c_iflag &= (~IXOFF);
		break;

	case MXF_232_HARDWARE_FLOW_CONTROL:
		/* Yes to hardware flow control. */
		tio.c_cflag &= (~CLOCAL);

		/* No to software flow control. */
		tio.c_iflag &= (~IXON);
		tio.c_iflag &= (~IXOFF);
		break;

	case MXF_232_SOFTWARE_FLOW_CONTROL:
		/* No to hardware flow control. */
		tio.c_cflag |= CLOCAL;

		/* Yes to software flow control. */
		tio.c_iflag |= IXON;
		tio.c_iflag |= IXOFF;
		break;

	case MXF_232_BOTH_FLOW_CONTROL:
		/* Yes to hardware flow control. */
		tio.c_cflag &= (~CLOCAL);

		/* Yes to software flow control. */
		tio.c_iflag |= IXON;
		tio.c_iflag |= IXOFF;
		break;

	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Unsupported TTY flow control = %d for TTY '%s'.", 
		rs232->flow_control, name);

		break;
	}

	/* Save a copy of the current settings so we can see if we
	 * got what we asked for.
	 */

	tio_copy = tio;

	/* Throw away any input data (noise) that has shown up since
	 * we started.
	 */

	if ( tcflush( tty_fd, TCIOFLUSH ) != 0 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Unable to clear input and output for TTY '%s' just before tcsetattr()",
			name);
	}

	/* Everything is set, so change the configuration of the TTY. */

	if ( tcsetattr( tty_fd, TCSANOW, &tio ) != 0 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
		"Unable to change current state of TTY '%s'.", name);
	}
		

	/* Check to see if the current setting match what we just 
	 * asked for.
	 */

	if ( tcgetattr( tty_fd, &tio ) != 0 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Unable to get current state of TTY '%s' before comparing changed settings.",
			name);
	}

	error = FALSE;

	if ( tio.c_iflag != tio_copy.c_iflag ) {
		error = TRUE;
		mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"tcsetattr() did not set c_iflag correctly, desired c_iflag = 0%o, "
	"actual c_iflag = 0%o", tio_copy.c_iflag, tio.c_iflag );
	}

	if ( tio.c_oflag != tio_copy.c_oflag ) {
		error = TRUE;
		mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"tcsetattr() did not set c_oflag correctly, desired c_oflag = 0%o, "
	"actual c_oflag = 0%o", tio_copy.c_oflag, tio.c_oflag );
	}

#if 0
	/* This is temporarily disabled.  Under Solaris 2 some high
	 * order bits of unknown nature are set in the returned value.
	 */
	if ( tio.c_cflag != tio_copy.c_cflag ) {
		error = TRUE;
		mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"tcsetattr() did not set c_cflag correctly, desired c_cflag = 0%o, "
	"actual c_cflag = 0%o", tio_copy.c_cflag, tio.c_cflag );
	}
#endif

	if ( tio.c_lflag != tio_copy.c_lflag ) {
		error = TRUE;
		mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"tcsetattr() did not set c_lflag correctly, desired c_lflag = 0%o, "
	"actual c_lflag = 0%o", tio_copy.c_lflag, tio.c_lflag );
	}

	for ( i = 0; i < NCCS; i++ ) {
		if ( tio.c_cc[i] != tio_copy.c_cc[i] ) {
			error = TRUE;
			mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"tcsetattr() did not set c_cc[%d] correctly, desired value = 0%o, "
	"actual value = 0%o", tio_copy.c_cc[i], tio.c_cc[i] );
			/* break; */
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

#endif /* USE_POSIX_TERMIOS */

#endif /* OS_UNIX */
