/*
 * Name:    mx_mcs.c
 *
 * Purpose: MX function library for multichannel scalers.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_alloc.h"
#include "mx_drivr.h"
#include "mx_mcs.h"

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_mcs_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_mcs_get_pointers( MX_RECORD *mcs_record,
			MX_MCS **mcs,
			MX_MCS_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_mcs_get_pointers()";

	if ( mcs_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The mcs_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( mcs_record->mx_class != MXC_MULTICHANNEL_SCALER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"The record '%s' passed by '%s' is not an MCS.",
			mcs_record->name, calling_fname );
	}

	if ( mcs != (MX_MCS **) NULL ) {
		*mcs = (MX_MCS *) (mcs_record->record_class_struct);

		if ( *mcs == (MX_MCS *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS pointer for record '%s' passed by '%s' is NULL",
				mcs_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_MCS_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_MCS_FUNCTION_LIST *)
				(mcs_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_MCS_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_MCS_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				mcs_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_mcs_finish_record_initialization( MX_RECORD *mcs_record )
{
	const char fname[] = "mx_mcs_finish_record_initialization()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	long i;
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( mcs->max_num_scalers == 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The multichannel scaler '%s' is set to have zero scaler channels.  "
	"This is an error since at least one scaler channel is required.",
			mcs_record->name );
	}

	if ( mcs->max_num_measurements == 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The multichannel scaler '%s' is set to have zero measurements.  "
	"This is an error since at least one measurement is required.",
			mcs_record->name );
	}

	mcs->num_scalers = mcs->max_num_scalers;
	mcs->num_measurements = mcs->max_num_measurements;

	mcs->readout_preference = MXF_MCS_PREFER_READ_SCALER;
	mcs->first_scaler_number = 0;

	mcs->timer_record = NULL;

	mcs->scaler_record_array = (MX_RECORD **) malloc( mcs->max_num_scalers
						* sizeof(MX_RECORD *) );

	if ( mcs->scaler_record_array == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a %ld element array of "
		"MCS scaler record pointers.", mcs->max_num_scalers );
	}

	for ( i = 0; i < mcs->max_num_scalers; i++ ) {
		mcs->scaler_record_array[i] = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_mcs_start( MX_RECORD *mcs_record )
{
	const char fname[] = "mx_mcs_start()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *start_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	start_fn = function_list->start;

	if ( start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for record '%s' is NULL.",
			mcs_record->name );
	}

	status = (*start_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_stop( MX_RECORD *mcs_record )
{
	const char fname[] = "mx_mcs_stop()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *stop_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	stop_fn = function_list->stop;

	if ( stop_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"stop function ptr for record '%s' is NULL.",
			mcs_record->name );
	}

	status = (*stop_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_clear( MX_RECORD *mcs_record )
{
	const char fname[] = "mx_mcs_clear()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *clear_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	clear_fn = function_list->clear;

	if ( clear_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"clear function ptr for record '%s' is NULL.",
			mcs_record->name );
	}

	status = (*clear_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_is_busy( MX_RECORD *mcs_record, int *busy )
{
	const char fname[] = "mx_mcs_is_busy()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *busy_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	busy_fn = function_list->busy;

	if ( busy_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"busy function ptr for record '%s' is NULL.",
			mcs_record->name );
	}

	status = (*busy_fn)( mcs );

	*busy = mcs->busy;

	MX_DEBUG( 1,("%s complete.  busy = %d", fname, mcs->busy));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_read_all( MX_RECORD *mcs_record )
{
	const char fname[] = "mx_mcs_read_all()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *read_all_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	read_all_fn = function_list->read_all;

	if ( read_all_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
	"Reading the MCS data as one big block is not supported for MCS '%s'.",
			mcs_record->name );
	}

	status = (*read_all_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_read_scaler( MX_RECORD *mcs_record, unsigned long scaler_index )
{
	const char fname[] = "mx_mcs_read_scaler()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *read_scaler_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s', scaler_index = %ld",
				fname, mcs_record->name, scaler_index));

	read_scaler_fn = function_list->read_scaler;

	if ( read_scaler_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Reading the MCS data one scaler channel at a time "
			"is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->scaler_index = scaler_index;

	status = (*read_scaler_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_read_measurement( MX_RECORD *mcs_record,
				unsigned long measurement_index )
{
	const char fname[] = "mx_mcs_read_measurement()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *read_measurement_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	read_measurement_fn = function_list->read_measurement;

	if ( read_measurement_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
			"Reading the MCS data one measurement at a time "
			"is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->measurement_index = measurement_index;

	status = (*read_measurement_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_get_mode( MX_RECORD *mcs_record, int *mode )
{
	const char fname[] = "mx_mcs_get_mode()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	get_parameter_fn = function_list->get_parameter;

	if ( get_parameter_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The get_parameter function is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->parameter_type = MXLV_MCS_MODE;

	status = (*get_parameter_fn)( mcs );

	*mode = mcs->mode;

	MX_DEBUG( 1,("%s complete.  mode = %d", fname, mcs->mode));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_set_mode( MX_RECORD *mcs_record, int mode )
{
	const char fname[] = "mx_mcs_set_mode()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *set_parameter_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'.  mode = %d",
			fname, mcs_record->name, mode));

	set_parameter_fn = function_list->set_parameter;

	if ( set_parameter_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The set_parameter function is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->parameter_type = MXLV_MCS_MODE;

	mcs->mode = mode;

	status = (*set_parameter_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_get_measurement_time( MX_RECORD *mcs_record, double *measurement_time )
{
	const char fname[] = "mx_mcs_get_measurement_time()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	get_parameter_fn = function_list->get_parameter;

	if ( get_parameter_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The get_parameter function is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->parameter_type = MXLV_MCS_MEASUREMENT_TIME;

	status = (*get_parameter_fn)( mcs );

	*measurement_time = mcs->measurement_time;

	MX_DEBUG( 1,("%s complete.  measurement_time = %g",
				fname, mcs->measurement_time));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_set_measurement_time( MX_RECORD *mcs_record, double measurement_time )
{
	const char fname[] = "mx_mcs_set_measurement_time()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *set_parameter_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'.  measurement_time = %g",
			fname, mcs_record->name, measurement_time));

	set_parameter_fn = function_list->set_parameter;

	if ( set_parameter_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The set_parameter function is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->parameter_type = MXLV_MCS_MEASUREMENT_TIME;

	mcs->measurement_time = measurement_time;

	status = (*set_parameter_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_get_num_measurements( MX_RECORD *mcs_record,
					unsigned long *num_measurements )
{
	const char fname[] = "mx_mcs_get_num_measurements()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *get_parameter_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'", fname, mcs_record->name));

	get_parameter_fn = function_list->get_parameter;

	if ( get_parameter_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The get_parameter function is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->parameter_type = MXLV_MCS_NUM_MEASUREMENTS;

	status = (*get_parameter_fn)( mcs );

	*num_measurements = mcs->num_measurements;

	MX_DEBUG( 1,("%s complete.  num_measurements = %ld",
				fname, mcs->num_measurements));

	return status;
}

MX_EXPORT mx_status_type
mx_mcs_set_num_measurements( MX_RECORD *mcs_record,
					unsigned long num_measurements )
{
	const char fname[] = "mx_mcs_set_num_measurements()";

	MX_MCS *mcs;
	MX_MCS_FUNCTION_LIST *function_list;
	mx_status_type ( *set_parameter_fn ) ( MX_MCS * );
	mx_status_type status;

	status = mx_mcs_get_pointers( mcs_record,
					&mcs, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 1,("%s invoked for MCS '%s'.  num_measurements = %ld",
			fname, mcs_record->name, num_measurements));

	set_parameter_fn = function_list->set_parameter;

	if ( set_parameter_fn == NULL ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The set_parameter function is not supported for MCS '%s'.",
			mcs_record->name );
	}

	mcs->parameter_type = MXLV_MCS_NUM_MEASUREMENTS;

	mcs->num_measurements = num_measurements;

	status = (*set_parameter_fn)( mcs );

	MX_DEBUG( 1,("%s complete.", fname));

	return status;
}

