/*
 * Name:     mx_motor.h
 *
 * Purpose:  Main header file for the MX motor class.
 *
 * Author:   William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#ifndef __MX_MOTOR_H__
#define __MX_MOTOR_H__

/* Include hardware and operating system dependent definitions and symbols. */

#include "mx_rec.h"

/* ------------------------------------------------------------------ */
#define MX_MOTOR_STEP_FUZZ    (1.0e-9)
#define MX_MOTOR_ANALOG_FUZZ  (1.0e-9)

/* Flag values for mx_motor_move_absolute() and similar routines. */

#define MXF_MTR_NONE               	0
#define MXF_MTR_SHOW_MOVE          	1
#define MXF_MTR_NOWAIT             	2
#define MXF_MTR_IGNORE_KEYBOARD    	4
#define MXF_MTR_IGNORE_BACKLASH    	8
#define MXF_MTR_GO_TO_BACKLASH_POSITION	16
#define MXF_MTR_ONLY_CHECK_LIMITS	32
#define MXF_MTR_SCAN_IN_PROGRESS	64

#define MX_MOTOR_NUM_ACCELERATION_PARAMS  4

/* Acceleration types. */

#define MXF_MTR_ACCEL_NONE			0
#define MXF_MTR_ACCEL_DIST_PER_SECOND_SQUARED	1
#define MXF_MTR_ACCEL_TIME			2

/* MX_MOTOR contains the data that is common to all motor types. */

typedef struct {
	MX_RECORD *record; /* Pointer to the MX_RECORD structure that points
			    * to this motor.
			    */

	union { long stepper;
		double analog;
	} raw_destination;

	union { long stepper;
		double analog;
	} raw_position;

	union { long stepper;
		double analog;
	} raw_backlash_correction;

	union { long stepper;
		double analog;
	} raw_negative_limit;

	union { long stepper;
		double analog;
	} raw_positive_limit;

	union { long stepper;
		double analog;
	} deadband;

	union { long stepper;
		double analog;
	} raw_set_position;

	double raw_minimum_speed_limit;
	double raw_maximum_speed_limit;

	double minimum_speed_limit;
	double maximum_speed_limit;

	int subclass;

	double scale;
	double offset;
	char   units[MXU_UNITS_NAME_LENGTH];

	double destination;
	double position;
	double set_position;
	double backlash_correction;
	double negative_limit;
	double positive_limit;

	double raw_speed;
	double raw_base_speed;
	double raw_maximum_speed;

	double speed;
	double base_speed;
	double maximum_speed;

	int acceleration_type;
	double acceleration_parameters[ MX_MOTOR_NUM_ACCELERATION_PARAMS ];

	int at_home_switch;

	int busy;
	int soft_abort;
	int immediate_abort;
	int negative_limit_hit;
	int positive_limit_hit;
	int home_search;
	int constant_velocity_move;
	int parameter_type;
} MX_MOTOR;

#define MXLV_MTR_BUSY				1001
#define MXLV_MTR_DESTINATION			1002
#define MXLV_MTR_POSITION			1003
#define MXLV_MTR_SET_POSITION			1004
#define MXLV_MTR_SOFT_ABORT			1005
#define MXLV_MTR_POSITIVE_LIMIT_HIT		1006
#define MXLV_MTR_NEGATIVE_LIMIT_HIT		1007
#define MXLV_MTR_HOME_SEARCH			1008
#define MXLV_MTR_CONSTANT_VELOCITY_MOVE		1009
#define MXLV_MTR_SPEED				1010
#define MXLV_MTR_BASE_SPEED			1011
#define MXLV_MTR_MAXIMUM_SPEED			1012
#define MXLV_MTR_ACCELERATION_TYPE		1013
#define MXLV_MTR_ACCELERATION_PARAMETERS	1014
#define MXLV_MTR_IS_AT_HOME_SWITCH		1015

#define MX_STEPPER_MOTOR_STANDARD_FIELDS \
  {-1, -1, "raw_destination", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_destination.stepper), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "raw_position", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_position.stepper), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_backlash_correction", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, \
					raw_backlash_correction.stepper), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_negative_limit", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_negative_limit.stepper),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_positive_limit", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_positive_limit.stepper),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "deadband", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, deadband.stepper),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_set_position", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_set_position.stepper), \
	{0}, NULL, 0}

#define MX_ANALOG_MOTOR_STANDARD_FIELDS \
  {-1, -1, "raw_destination", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_destination.analog), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "raw_position", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_position.analog), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_backlash_correction", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, \
					raw_backlash_correction.analog), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_negative_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_negative_limit.analog),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_positive_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_positive_limit.analog),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "deadband", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, deadband.analog),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_set_position", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_set_position.analog), \
	{0}, NULL, 0}

#define MX_MOTOR_STANDARD_FIELDS \
  {-1, -1, "raw_minimum_speed_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_minimum_speed_limit),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_maximum_speed_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_maximum_speed_limit),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "minimum_speed_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, minimum_speed_limit),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "maximum_speed_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, maximum_speed_limit),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "subclass", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, subclass), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "scale", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, scale), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "offset", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, offset), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "units", MXFT_STRING, NULL, 1, {MXU_UNITS_NAME_LENGTH}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, units), \
	{sizeof(char)}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {MXLV_MTR_DESTINATION, -1, "destination", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, destination), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_POSITION, -1, "position", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, position), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_SET_POSITION, -1, "set_position", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, set_position), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "backlash_correction", MXFT_DOUBLE, NULL, 0, {0},\
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, backlash_correction), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "negative_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, negative_limit),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "positive_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, positive_limit),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "raw_speed", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_speed),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "raw_base_speed", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_base_speed),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "raw_maximum_speed", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_maximum_speed),\
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_SPEED, -1, "speed", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, speed),\
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_BASE_SPEED, -1, "base_speed", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, base_speed),\
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_MAXIMUM_SPEED, -1, "maximum_speed", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, maximum_speed),\
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_ACCELERATION_TYPE, -1, \
		"acceleration_type", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, acceleration_type), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_ACCELERATION_PARAMETERS, -1, "acceleration_parameters", \
		MXFT_INT, NULL, 1, {MX_MOTOR_NUM_ACCELERATION_PARAMS }, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, acceleration_type), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_BUSY, -1, "busy", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, busy), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_SOFT_ABORT, -1, "soft_abort", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, soft_abort), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "immediate_abort", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, immediate_abort), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_POSITIVE_LIMIT_HIT, -1, "negative_limit_hit", MXFT_INT, \
	NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, negative_limit_hit), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_NEGATIVE_LIMIT_HIT, -1, "positive_limit_hit", MXFT_INT, \
	NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, positive_limit_hit), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_HOME_SEARCH, -1, "home_search", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, home_search), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_CONSTANT_VELOCITY_MOVE, -1, "constant_velocity_move", \
					MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, constant_velocity_move), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "parameter_type", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, parameter_type), \
	{0}, NULL, 0}

/*
 * The structure type MX_MOTOR_FUNCTION_LIST contains a list of
 * pointers to functions that vary from motor type to motor type.
 * By abstracting away the individual motor operations in this
 * fashion, it is possible to minimize device dependencies in
 * the main motor source code.
 */

typedef struct {
	mx_status_type ( *motor_is_busy )( MX_MOTOR *motor );
	mx_status_type ( *move_absolute )( MX_MOTOR *motor );
	mx_status_type ( *get_position )( MX_MOTOR *motor );
	mx_status_type ( *set_position )( MX_MOTOR *motor );
	mx_status_type ( *soft_abort )( MX_MOTOR *motor );
	mx_status_type ( *immediate_abort )( MX_MOTOR *motor );
	mx_status_type ( *positive_limit_hit )( MX_MOTOR *motor );
	mx_status_type ( *negative_limit_hit )( MX_MOTOR *motor );
	mx_status_type ( *find_home_position )( MX_MOTOR *motor );
	mx_status_type ( *constant_velocity_move )( MX_MOTOR *motor );
	mx_status_type ( *get_parameter )( MX_MOTOR *motor );
	mx_status_type ( *set_parameter )( MX_MOTOR *motor );
} MX_MOTOR_FUNCTION_LIST;

typedef mx_status_type (*MX_MOTOR_MOVE_REPORT_FUNCTION)(int, int, MX_RECORD **);

/* ============== Public functions and symbols . ================= */

MX_API_PRIVATE mx_status_type mx_motor_finish_record_initialization(
						MX_RECORD *motor_record );

MX_API_PRIVATE mx_status_type mx_motor_get_pointers( MX_RECORD *motor_record,
		MX_MOTOR **motor, MX_MOTOR_FUNCTION_LIST **function_list_ptr,
		const char *calling_fname );

MX_API mx_status_type mx_motor_is_busy( MX_RECORD *motor_record, int *busy );

MX_API mx_status_type mx_motor_soft_abort( MX_RECORD *motor_record );

MX_API mx_status_type mx_motor_immediate_abort( MX_RECORD *motor_record );

/* === Move by engineering units functions. === */

#define mx_motor_move_relative( r, rp, flags ) \
		mx_motor_move_relative_with_report( (r), (rp), NULL, (flags) );

#define mx_motor_move_absolute( r, p, flags ) \
		mx_motor_move_absolute_with_report( (r), (p), NULL, (flags) );

#define mx_motor_array_move_absolute( nr, mra, pa, flags ) \
		mx_motor_array_move_absolute_with_report( \
				(nr), (mra), (pa), NULL, (flags) );

MX_API mx_status_type mx_motor_move_relative_with_report(
			MX_RECORD *motor_record,
			double relative_position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags);

MX_API mx_status_type mx_motor_move_absolute_with_report(
			MX_RECORD *motor_record,
			double position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags);

MX_API mx_status_type mx_motor_array_move_absolute_with_report(
			int num_motor_records,
			MX_RECORD **motor_record_array,
			double *position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags );

MX_API mx_status_type mx_motor_array_move_segment_with_report(
			int num_motor_records,
			MX_RECORD **motor_record_array,
			double *position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags );

MX_API mx_status_type mx_wait_for_motor_stop(
			MX_RECORD *motor_record,
			int flags );

MX_API mx_status_type mx_wait_for_motor_array_stop(
			int num_motor_records,
			MX_RECORD **motor_record_array,
			int flags );

MX_API mx_status_type mx_motor_get_position( MX_RECORD *motor_record,
							double *position );
MX_API mx_status_type mx_motor_set_position( MX_RECORD *motor_record, 
							double position );

MX_API mx_status_type mx_motor_positive_limit_hit(
				MX_RECORD *motor_record, int *limit_hit );

MX_API mx_status_type mx_motor_negative_limit_hit(
				MX_RECORD *motor_record, int *limit_hit );

MX_API mx_status_type mx_motor_find_home_position(
				MX_RECORD *motor_record, int direction );

MX_API mx_status_type mx_motor_zero_position_value( MX_RECORD *motor_record );

MX_API mx_status_type mx_motor_constant_velocity_move(
				MX_RECORD *motor_record, int direction );

MX_API mx_status_type mx_motor_get_parameter(
				MX_RECORD *motor_record, int parameter_type );

MX_API mx_status_type mx_motor_set_parameter(
				MX_RECORD *motor_record, int parameter_type );

MX_API mx_status_type mx_motor_get_speed( MX_RECORD *motor_record,
						double *speed );

MX_API mx_status_type mx_motor_set_speed( MX_RECORD *motor_record,
						double speed );

MX_API mx_status_type mx_motor_get_base_speed( MX_RECORD *motor_record,
						double *base_speed );

MX_API mx_status_type mx_motor_set_base_speed( MX_RECORD *motor_record,
						double base_speed );

MX_API mx_status_type mx_motor_get_maximum_speed( MX_RECORD *motor_record,
						double *maximum_speed );

MX_API mx_status_type mx_motor_set_maximum_speed( MX_RECORD *motor_record,
						double maximum_speed );

MX_API mx_status_type mx_motor_get_acceleration_type( MX_RECORD *motor_record,
						int *acceleration_type );

MX_API mx_status_type mx_motor_get_acceleration_parameters(
				MX_RECORD *motor_record,
				double *acceleration_parameter_array );

MX_API mx_status_type mx_motor_set_acceleration_parameters(
				MX_RECORD *motor_record,
				double *acceleration_parameter_array );

MX_API mx_status_type mx_motor_is_at_home_switch(
				MX_RECORD *motor_record, int *at_home_switch );

MX_API mx_status_type mx_is_motor_position_between_software_limits(
			MX_RECORD *motor_record, double position,
			int *result_flag, int generate_error_message );

/* === Move by steps functions. (MXC_MTR_STEPPER) === */

MX_API mx_status_type mx_motor_move_relative_steps_with_report(
			MX_RECORD *motor_record,
			long motor_steps,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags);

MX_API mx_status_type mx_motor_move_absolute_steps_with_report(
			MX_RECORD *motor_record,
			long motor_steps,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags);

MX_API mx_status_type mx_motor_get_position_steps(
				MX_RECORD *motor_record, long *motor_steps );
MX_API mx_status_type mx_motor_set_position_steps(
				MX_RECORD *motor_record, long motor_steps );

MX_API mx_status_type mx_motor_steps_to_go(
				MX_RECORD *motor_record, long *steps_to_go );

MX_API mx_status_type
		mx_is_stepper_motor_position_between_software_limits(
			MX_RECORD *motor_record, long step_position,
			int *result_flag, int generate_error_message );

/* === Move analog functions. (MXC_MTR_ANALOG) === */

MX_API mx_status_type mx_motor_move_absolute_analog_with_report(
			MX_RECORD *motor_record,
			double motor_position,
			MX_MOTOR_MOVE_REPORT_FUNCTION move_report_fn,
			int flags);

MX_API mx_status_type mx_motor_get_position_analog(
				MX_RECORD *motor_record, double *position );
MX_API mx_status_type mx_motor_set_position_analog(
				MX_RECORD *motor_record, double position );

MX_API mx_status_type
		mx_is_analog_motor_position_between_software_limits(
			MX_RECORD *motor_record, double analog_position,
			int *result_flag, int generate_error_message );

/* === */

extern MX_RECORD_FUNCTION_LIST mx_motor_record_function_list;

#endif /* __MX_MOTOR_H__ */
