/*
 * Name:    sq_struc.c
 *
 * Purpose: Driver for quick scans that use a Struck MCS via EPICS.
 *
 * Author:  William Lavender
 *
 *--------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_alloc.h"
#include "mx_scan.h"
#include "mx_scqui.h"
#include "sq_struc.h"
#include "m_time.h"
#include "d_pmactc.h"
#include "d_epmtr.h"
#include "d_epmcs.h"
#include "d_mcsscl.h"
#include "d_mcstim.h"

/* EPICS include files. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

MX_RECORD_FUNCTION_LIST mxs_struck_quick_scan_record_function_list = {
	mxs_struck_quick_scan_initialize_type,
	mxs_struck_quick_scan_create_record_structures,
	mxs_struck_quick_scan_finish_record_initialization,
	mxs_struck_quick_scan_delete_record,
	mx_quick_scan_print_scan_structure,
	mxs_struck_quick_scan_dummy_function,
	mxs_struck_quick_scan_dummy_function,
	mxs_struck_quick_scan_dummy_function,
	mxs_struck_quick_scan_dummy_function
};

MX_SCAN_FUNCTION_LIST mxs_struck_quick_scan_scan_function_list = {
	mxs_struck_quick_scan_prepare_for_scan_start,
	mxs_struck_quick_scan_execute_scan_body,
	mxs_struck_quick_scan_cleanup_after_scan_end
};

MX_RECORD_FIELD_DEFAULTS mxs_struck_quick_scan_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_SCAN_STANDARD_FIELDS,
	MX_QUICK_SCAN_STANDARD_FIELDS
};

long mxs_struck_quick_scan_num_record_fields
			= sizeof( mxs_struck_quick_scan_defaults )
			/ sizeof( mxs_struck_quick_scan_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxs_struck_quick_scan_def_ptr
			= &mxs_struck_quick_scan_defaults[0];

#define EPICS_NAME_LENGTH	41

#define FREE_ARRAYS mxs_struck_quick_scan_free_arrays(scan,struck_quick_scan)

static void
mxs_struck_quick_scan_free_arrays( MX_SCAN *scan,
				MX_STRUCK_QUICK_SCAN *struck_quick_scan )
{
	long i;

	if ( scan == NULL )
		return;

	if ( struck_quick_scan == NULL )
		return;

	if ( struck_quick_scan->motor_position_array != NULL ) {
	    for ( i = 0; i < scan->num_motors; i++ ) {
		if ( struck_quick_scan->motor_position_array[i] != NULL ) {
		    free( struck_quick_scan->motor_position_array[i] );

		    struck_quick_scan->motor_position_array[i] = NULL;
		}
	    }
	}
}

static mx_status_type
mxs_struck_quick_scan_get_pointers( MX_SCAN *scan,
				MX_QUICK_SCAN **quick_scan,
				MX_STRUCK_QUICK_SCAN **struck_quick_scan,
				const char *calling_fname )
{
	const char fname[] = "mxs_struck_quick_scan_get_pointers()";

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_SCAN pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( quick_scan == (MX_QUICK_SCAN **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_QUICK_SCAN pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( struck_quick_scan == (MX_STRUCK_QUICK_SCAN **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_STRUCK_QUICK_SCAN pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*quick_scan = (MX_QUICK_SCAN *) (scan->record->record_class_struct);

	if ( *quick_scan == (MX_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_QUICK_SCAN pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	*struck_quick_scan = (MX_STRUCK_QUICK_SCAN *)
					scan->record->record_type_struct;

	if ( *struck_quick_scan == (MX_STRUCK_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_STRUCK_QUICK_SCAN pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_initialize_type( long type )
{
	return mx_quick_scan_initialize_type( type );
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_create_record_structures( MX_RECORD *record )
{
	const char fname[] =
		"mxs_struck_quick_scan_create_record_structures()";

	MX_SCAN *scan;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	int i;

	/* Allocate memory for the necessary structures. */

	scan = (MX_SCAN *) malloc( sizeof(MX_SCAN) );

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCAN structure." );
	}

	quick_scan = (MX_QUICK_SCAN *) malloc( sizeof(MX_QUICK_SCAN) );

	if ( quick_scan == (MX_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_QUICK_SCAN structure." );
	}

	struck_quick_scan = (MX_STRUCK_QUICK_SCAN *)
				malloc( sizeof(MX_STRUCK_QUICK_SCAN) );

	if ( struck_quick_scan == (MX_STRUCK_QUICK_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_STRUCK_QUICK_SCAN structure.");
	}

	/* Now set up the necessary pointers. */

	record->record_superclass_struct = scan;
	record->record_class_struct = quick_scan;
	record->record_type_struct = struck_quick_scan;

	scan->record = record;

	scan->measurement.scan = scan;
	scan->datafile.scan = scan;
	scan->plot.scan = scan;

	quick_scan->scan = scan;

	record->superclass_specific_function_list
				= &mxs_struck_quick_scan_scan_function_list;

	record->class_specific_function_list = NULL;

	for ( i = 0; i < MXS_SQ_MAX_MOTORS; i++ ) {
		struck_quick_scan->motor_position_array[i] = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] =
		"mxs_struck_quick_scan_finish_record_initialization()";

	MX_SCAN *scan;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	MX_RECORD *input_device_record;
	MX_MCS_SCALER *mcs_scaler;
	MX_RECORD *mcs_record;
	MX_MCS *mcs;
	long i;
	mx_status_type status;

	scan = (MX_SCAN *) record->record_superclass_struct;

	status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	input_device_record = ( scan->input_device_array )[0];

	if ( input_device_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The first input device record pointer for scan '%s' is NULL.",
			record->name );
	}

	mcs_scaler = (MX_MCS_SCALER *) input_device_record->record_type_struct;

	if ( mcs_scaler == (MX_MCS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS_SCALER pointer for input device '%s' is NULL.",
			input_device_record->name );
	}

	mcs_record = mcs_scaler->mcs_record;

	if ( mcs_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mcs record pointer for MCS scaler '%s' is NULL.",
			input_device_record->name );
	}

	mcs = (MX_MCS *) mcs_record->record_class_struct;

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS pointer for MCS record '%s' is NULL.",
			mcs_record->name );
	}

	if ( ( scan->num_motors < 0 )
	  || ( scan->num_motors > MXS_SQ_MAX_MOTORS ) )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal num_motors value %ld for scan '%s'.  "
		"The allowed range is (0-%ld)",
			scan->num_motors, record->name,
			MXS_SQ_MAX_MOTORS );
	}

	if ( ( scan->num_input_devices < 0 )
	  || ( scan->num_input_devices > mcs->max_num_scalers ) )
	{
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal num_input_devices value %ld for scan '%s'.  "
		"The allowed range is (0-%ld)",
			scan->num_input_devices, record->name,
			mcs->max_num_scalers );
	}

	for ( i = 0; i < scan->num_motors; i++ ) {
		scan->motor_is_independent_variable[i] = TRUE;
	}

	return mx_scan_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_dummy_function( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_struck_quick_scan_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mxs_struck_quick_scan_prepare_for_scan_start()";

	MX_RECORD *mcs_record;
	MX_MCS_SCALER *mcs_scaler;
	MX_RECORD *input_device_record;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	MX_MEASUREMENT_PRESET_TIME *preset_time_struct;
	long i;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Verify that all the input devices are MCS scaler records. */

	mcs_record = NULL;

	for ( i = 0; i < scan->num_input_devices; i++ ) {

		input_device_record = ( scan->input_device_array )[i];

		if ( (input_device_record->mx_superclass != MXR_DEVICE)
		  || (input_device_record->mx_class != MXC_SCALER)
		  || (input_device_record->mx_type != MXT_SCL_MCS) )
		{
			return mx_error( MXE_TYPE_MISMATCH, fname,
	"Input device '%s' is not an MCS channel.  Only MCS scaler records "
	"are allowed as input devices for an MCS quick scan.",
				input_device_record->name );
		}

		mcs_scaler = (MX_MCS_SCALER *)
				input_device_record->record_type_struct;

		if ( mcs_scaler == (MX_MCS_SCALER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MCS scaler pointer for input device '%s' is NULL.",
				input_device_record->name );
		}
		if ( mcs_scaler->mcs_record == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The mcs_record pointer for MCS scaler '%s' is NULL.",
				input_device_record->name );
		}

		if ( mcs_record == NULL ) {
			mcs_record = mcs_scaler->mcs_record;
		} else if ( mcs_record != mcs_scaler->mcs_record ) {

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The MCS scaler channels for this scan do not all belong to the "
	"same multichannel scaler.  This is not permitted, so you will "
	"need to change this scan's list of input devices." );
		}
	}

	/* Verify that the timer is an MCS timer record. */

	preset_time_struct = (MX_MEASUREMENT_PRESET_TIME *)
				scan->measurement.measurement_type_struct;

	if ( preset_time_struct == (MX_MEASUREMENT_PRESET_TIME *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MEASUREMENT_PRESET_TIME pointer for scan '%s' is NULL.",
			scan->record->name );
	}
	if ( preset_time_struct->timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"timer_record pointer in preset_time_struct for scan '%s' is NULL.",
			scan->record->name );
	}
	if ( preset_time_struct->timer_record->mx_type != MXT_TIM_MCS ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The timer record '%s' for scan '%s' is not an MCS timer record.  "
	"You may only use MCS timers for MCS quick scans.",
			preset_time_struct->timer_record->name,
			scan->record->name );
	}

	/* Allocate memory for the motor position array. */

	for ( i = 0; i < scan->num_motors; i++ ) {

		struck_quick_scan->motor_position_array[i] = (double *)
			malloc(quick_scan->num_measurements * sizeof(double));

		if ( struck_quick_scan->motor_position_array[i] == NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Ran out of memory trying to allocate an %ld array "
			"of motor positions for motor %ld.",
				quick_scan->num_measurements, i );
		}
	}

	/* Initialize the datafile and plotting support. */

	status = mx_standard_prepare_for_scan_start( scan );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	if ( mx_plotting_is_enabled( scan->record ) ) {
		status = mx_plot_start_plot_section( &(scan->plot) );

		if ( status.code != MXE_SUCCESS ) {
			FREE_ARRAYS;
			return status;
		}
	}

	/**** Put the MCS into preset time mode. ****/

	status = mx_mcs_set_mode( mcs_record, MXM_PRESET_TIME );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	/**** Set the measurement time per point. ****/

	if ( scan->measurement.type != MXM_PRESET_TIME ) {
		FREE_ARRAYS;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Quick scans only support preset time mode, but the scan '%s' "
		"is not configured as a preset time scan.",
			scan->record->name );
	}

	status = mx_mcs_set_measurement_time( mcs_record,
				preset_time_struct->integration_time );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	/**** Set the number of measurements. ****/

	status = mx_mcs_set_num_measurements( mcs_record,
				(unsigned long) quick_scan->num_measurements );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	/**** Send the motors to the start position for the scan. ****/

	mx_info("Moving to the start position.");

	status = mx_motor_array_move_absolute(
			scan->num_motors,
			scan->motor_record_array,
			quick_scan->start_position,
			0 );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	/**** Wait for the motors to get to their start positions. ****/

	status = mx_wait_for_motor_array_stop( scan->num_motors,
				scan->motor_record_array, 0 );

	if ( status.code != MXE_SUCCESS ) {
		FREE_ARRAYS;
		return status;
	}

	mx_info("All motors are at the start position.");

	/**** Set the motor speeds for this quick scan. ****/

	for ( i = 0; i < scan->num_motors; i++ ) {

		status = mx_quick_scan_set_new_motor_speed( quick_scan, i );

		if ( status.code != MXE_SUCCESS ) {
			FREE_ARRAYS;
			return status;
		}
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/* Do not use FREE_ARRAYS in the execute_scan_body() function since
 * the cleanup_after_scan_end() function will attempt to use the arrays
 * regardless of what error code this routine returns.
 */

MX_EXPORT mx_status_type
mxs_struck_quick_scan_execute_scan_body( MX_SCAN *scan )
{
	const char fname[] = "mxs_struck_quick_scan_execute_scan_body()";

	MX_RECORD *mcs_record;
	MX_RECORD *input_device_record;
	MX_MCS_SCALER *mcs_scaler;
	MX_MCS *mcs;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	int direction, busy;
	long i, j;
	unsigned long measurement_milliseconds;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	mx_status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	input_device_record = ( scan->input_device_array )[0];

	mcs_scaler = (MX_MCS_SCALER *) input_device_record->record_type_struct;

	mcs_record = mcs_scaler->mcs_record;

	mcs = (MX_MCS *) mcs_record->record_class_struct;

	measurement_milliseconds = mx_round( 1000.0 *
			mx_quick_scan_get_measurement_time(quick_scan) );

	if ( measurement_milliseconds < 1L ) {
		measurement_milliseconds = 1L;
	}

	/* Start the MCS. */

	mx_info("Starting the MCS.");

	mx_status = mx_mcs_start( mcs_record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Wait several measurement times before starting the motors.
	 *
	 * This allows the MCS to record several points at the start
	 * position.  Since the absolute position of the motor at the
	 * start position is known from information provided by the
	 * motor controller, it is possible to calculate the motor
	 * position for each point in the scan by using the incremental
	 * encoder counts as an offset from the know absolute starting
	 * position.
	 */

	mx_msleep( 5L * measurement_milliseconds );

	/* Start the motors. */

	mx_info("Starting the motors.");

	for ( i = 0; i < scan->num_motors; i++ ) {

		if ( quick_scan->end_position[i]
			>= quick_scan->start_position[i] )
		{
			direction = 1;
		} else {
			direction = -1;
		}

		mx_status = mx_motor_constant_velocity_move(
				scan->motor_record_array[i], direction );

		if ( mx_status.code != MXE_SUCCESS ) {
			for ( j = 0; j < i; j++ ) {
				(void) mx_motor_soft_abort(
					scan->motor_record_array[i] );
			}
			return mx_status;
		}
	}

	mx_info("Quick scan is in progress...");

	/* Wait for the counting to finish. */

	do {
		mx_status = mx_mcs_is_busy( mcs_record, &busy );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		mx_msleep(100);

	} while ( busy );

	mx_info("Quick scan complete.  Stopping the motors.");

	/* Stop the motors. */

	for ( i = 0; i < scan->num_motors; i++ ) {
		(void) mx_motor_soft_abort( scan->motor_record_array[i] );
	}

	mx_info("Reading out the MCS.");

	/*  Get the data from the MCS channels. */

	for ( i = 0; i < mcs->max_num_scalers; i++ ) {

		mx_status = mx_mcs_read_scaler( mcs_record, i );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	MX_DEBUG( 2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

static mx_status_type
mxs_struck_quick_scan_compute_motor_positions(
		MX_SCAN *scan,
		MX_QUICK_SCAN *quick_scan,
		MX_STRUCK_QUICK_SCAN *struck_quick_scan,
		MX_MCS *mcs )
{
	const char fname[] = "mxs_struck_quick_scan_compute_motor_positions()";

	MX_RECORD *encoder_channels_record;
	double **motor_position_array;
	double step_size;
	long i, j;

	MX_DEBUG(-2,("%s invoked.", fname ));

	if ( scan->num_motors <= 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	motor_position_array = struck_quick_scan->motor_position_array;

	/* Set the first motor position for each motor to the start position.
	 *
	 * This logic ASSUMES that at least the first MCS measurement was
	 * taken before the motors were commanded to start moving.
	 */

	for ( j = 0; j < scan->num_motors; j++ ) {
		motor_position_array[j][0] = quick_scan->start_position[j];
	}

	/* If there was only one measurement (not likely), we are done. */

	if ( quick_scan->num_measurements <= 1 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* Find the database record containing the encoder channel numbers. */

	encoder_channels_record = NULL;		/* This is temporary. */

	/* Now compute the motor positions. */

	if ( encoder_channels_record == (MX_RECORD *) NULL ) {

		/* Use dead reckoning. */

		for ( j = 0; j < scan->num_motors; j++ ) {
			step_size = ( quick_scan->end_position[j]
				- quick_scan->start_position[j] )
				/ (double) ( quick_scan->num_measurements - 1 );

			for ( i = 1; i < quick_scan->num_measurements; i++ ) {
				motor_position_array[j][i] =
					quick_scan->start_position[j]
					+ step_size * (double) i;
			}
		}


	} else {
		/* Use actual encoder counts. */

		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"Reading out encoder up and down bits is not yet "
			"implemented in this version of the code" );
	}

	return MX_SUCCESSFUL_RESULT;
}

/* Please note that most error conditions are ignored in the
 * cleanup_after_scan_end() function since it is important that
 * the data be saved to disk in spite of any extraneous errors
 * that may occur.
 */

MX_EXPORT mx_status_type
mxs_struck_quick_scan_cleanup_after_scan_end( MX_SCAN *scan )
{
	const char fname[] = "mxs_struck_quick_scan_cleanup_after_scan_end()";

	MX_RECORD *mcs_record;
	MX_RECORD *input_device_record;
	MX_MCS *mcs;
	MX_MCS_SCALER *mcs_scaler;
	MX_QUICK_SCAN *quick_scan;
	MX_STRUCK_QUICK_SCAN *struck_quick_scan;
	unsigned long **data_array;
	double motor_positions[ MXS_SQ_MAX_MOTORS ];
	long *data_values;
	long i, j, scaler_index;
	mx_status_type mx_status;

	MX_DEBUG( 2,("%s invoked.", fname));

	mx_status = mxs_struck_quick_scan_get_pointers( scan,
			&quick_scan, &struck_quick_scan, fname );

	if ( mx_status.code != MXE_SUCCESS ) {

		/* Can't go any farther if this fails, so we abort. */

		FREE_ARRAYS;
		return mx_status;
	}

	MX_DEBUG( 2,("%s: restoring motor speeds.", fname));

	/* Restore the old motor speeds for this quick scan. */

	for ( i = 0; i < scan->num_motors; i++ ) {

		/* Errors in restoring the motor speed are ignored. */

		(void) mx_quick_scan_restore_old_motor_speed( quick_scan, i );
	}

	/* Copy the data in the scan data arrays to the datafile
	 * and send another copy to the plotting program.
	 */

	MX_DEBUG( 1,("%s: scan = %p", fname, scan));
	MX_DEBUG( 1,("%s: scan->input_device_array = %p",
				fname, scan->input_device_array));

	input_device_record = (scan->input_device_array)[0];

	MX_DEBUG( 1,("%s: input_device_record = %p",
				fname, input_device_record));

	mcs_scaler = (MX_MCS_SCALER *)input_device_record->record_type_struct;

	MX_DEBUG( 1,("%s: mcs_scaler = %p", fname, mcs_scaler));

	mcs_record = mcs_scaler->mcs_record;

	MX_DEBUG( 1,("%s: mcs_record = %p", fname, mcs_record));
	MX_DEBUG( 1,("%s: mcs_record->name = '%s'", fname, mcs_record->name));

	mcs = (MX_MCS *) mcs_record->record_class_struct;

	MX_DEBUG( 1,("%s: mcs = %p", fname, mcs));

	/* Calculate the positions of the first motor depending on either
	 * incremental encoder counts sent as UP counts and DOWN counts
	 * to a pair of scaler channels or by using dead reckoning.
	 */

	mx_status = mxs_struck_quick_scan_compute_motor_positions(
			scan, quick_scan, struck_quick_scan, mcs );

	if ( mx_status.code != MXE_SUCCESS ) {
		return mx_status;
	}

	/* Allocate an array to contain the scaler channel values for
	 * one measurement.
	 */

	data_values = (long *) malloc( mcs->max_num_scalers * sizeof(long) );

	MX_DEBUG( 1,("%s: data_values = %p", fname, data_values));

	if ( data_values == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Cannot allocate a %ld element array of scaler values.",
			mcs->max_num_scalers );
	}

	MX_DEBUG( 1,("%s: mcs->data_array = %p", fname, mcs->data_array));

	data_array = mcs->data_array;

	MX_DEBUG( 1,("%s: data_array = %p", fname, data_array));

	for ( i = 0; i < quick_scan->num_measurements; i++ ) {

		MX_DEBUG( 1,("%s: i = %d", fname, i));

		for ( j = 0; j < scan->num_motors; j++ ) {
			motor_positions[j] = 
				struck_quick_scan->motor_position_array[j][i];
		}
		for ( j = 0; j < scan->num_input_devices; j++ ) {

			MX_DEBUG( 1,("%s:    j = %d", fname, j));

			input_device_record = (scan->input_device_array)[j];

			MX_DEBUG( 1,("%s:    input_device_record = %p",
					fname, input_device_record));

			MX_DEBUG( 1,("%s:    input_device_record = '%s'",
					fname, input_device_record->name));

			mcs_scaler = (MX_MCS_SCALER *)
				input_device_record->record_type_struct;

			MX_DEBUG( 1,("%s:    mcs_scaler = %p",
					fname, mcs_scaler));

			MX_DEBUG( 1,("%s:    mcs_scaler->scaler_number = %ld",
					fname, mcs_scaler->scaler_number));

			scaler_index = mcs_scaler->scaler_number
						- mcs->first_scaler_number;

			MX_DEBUG( 1,("%s:    scaler_index = %ld",
					fname, scaler_index));

			MX_DEBUG( 1,("%s:    data_values = %p, data_array = %p",
					fname, data_values, data_array));

			MX_DEBUG( 1,("%s:    data_array[%d] = %p",
				fname, scaler_index,
				data_array[ scaler_index ] ));

			MX_DEBUG( 1,("%s:    data_array[%d][%d] = %ld",
				fname, mcs_scaler->scaler_number, i,
				data_array[ scaler_index ][i] ));

			data_values[j]
				= data_array[ scaler_index ][i];
		}

#if 1
		for ( j = 0; j < scan->num_motors; j++ ) {
			fprintf(stderr," %-8g", motor_positions[j]);
		}
		fprintf(stderr," : ");
		for ( j = 0; j < scan->num_input_devices; j++ ) {
			fprintf(stderr," %ld", data_values[j]);
		}
		fprintf(stderr,"\n");
#endif

		MX_DEBUG( 2,("%s: Copying measurement %ld to data file.",
				fname, i));

		mx_status = mx_add_array_to_datafile( &(scan->datafile),
			MXFT_DOUBLE, scan->num_motors, motor_positions,
			MXFT_LONG, scan->num_input_devices, data_values );

		if ( mx_status.code != MXE_SUCCESS ) {

			/* If we cannot write the data to the datafile,
			 * all is lost, so we abort.
			 */

			FREE_ARRAYS;
			free(data_values);
			return mx_status;
		}

		if ( mx_plotting_is_enabled( scan->record ) ) {
			MX_DEBUG( 2,(
			    "%s: Adding measurement %ld to plot buffer.",
				fname, i));

			/* Failing to update the plot correctly is not
			 * a reason to abort since that would interrupt
			 * the writing of the datafile.
			 */

			(void) mx_add_array_to_plot_buffer( &(scan->plot),
			    MXFT_DOUBLE, scan->num_motors, motor_positions,
			    MXFT_LONG, scan->num_input_devices, data_values );
		}
	}

	if ( mx_plotting_is_enabled( scan->record ) ) {
		MX_DEBUG( 2,("%s: displaying the plot.", fname));

		/* Attempt to display the plot, but do not abort on an error.*/

		(void) mx_display_plot( &(scan->plot) );
	}

	/* Close the datafile and shut down the plot. */

	MX_DEBUG( 2,("%s: Invoking standard cleanup function.", fname));

	mx_status = mx_standard_cleanup_after_scan_end( scan );

	/* Free all of the data arrays that were allocated in the routine
	 * mxs_struck_quick_scan_prepare_for_scan_start().
	 */

	MX_DEBUG( 2,("%s: freeing the arrays.", fname));

	FREE_ARRAYS;
	free(data_values);

	MX_DEBUG( 2,("%s complete.", fname));

	return mx_status;
}

#endif /* HAVE_EPICS */

