/*
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mdisp.c  --  Motor display motor move function. */

#include <stdio.h>

#include "motor.h"
#include "mx_key.h"

mx_status_type
motor_move_report_function( int flags,
				int num_motors,
				MX_RECORD **motor_record )
{
	const char fname[] = "motor_move_report_function()";

	MX_MOTOR *motor;
	int i;
	long current_steps;
	double current_position;
	int motors_busy, busy;
	int exit_loop, return_code;
	int positive_limit_hit, negative_limit_hit, abort_motors;
	mx_status_type mx_status;

	/* Don't bother doing anything if the MXF_MTR_NOWAIT
	 * flag is set.
	 */

	if ( flags & MXF_MTR_NOWAIT ) {
		return MX_SUCCESSFUL_RESULT;
	}

	if ( num_motors <= 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Invoked with an invalid number of motors = %d", num_motors );
	}

	/* Print the header (if necessary). */

	if ( flags & MXF_MTR_SHOW_MOVE ) {
		if ( num_motors == 1 ) {
			fprintf( output,
				"   Position    Steps");
		} else {
			for ( i = 0; i < num_motors; i++ ) {
				fprintf( output, 
				"   Pos(%d)   Steps(%d) ",
					i, i );
			}
		}
		fprintf( output, "\n" );
	}

	/* Begin motor display loop. */

	return_code = SUCCESS;
	exit_loop = FALSE;
	positive_limit_hit = FALSE;
	negative_limit_hit = FALSE;
	abort_motors = FALSE;

	while ( exit_loop == FALSE ) {
		motors_busy = FALSE;

		for ( i = 0; i < num_motors; i++ ) {

			motor = (MX_MOTOR *)
				(motor_record[i]->record_class_struct);

			/* Get current motion status. */

			mx_status = mx_motor_is_busy(motor_record[i], &busy);

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			if ( busy ) {
				motors_busy = TRUE;
			}

			/* Get current position. */

			switch ( motor->subclass ) {
			case MXC_MTR_STEPPER:
				mx_status = mx_motor_get_position_steps(
					motor_record[i], &current_steps );

				if ( mx_status.code != MXE_SUCCESS )
					return mx_status;

				current_position = motor->offset
				    + (motor->scale) * (double) current_steps;

				break;

			case MXC_MTR_ANALOG:
				mx_status = mx_motor_get_position(
					motor_record[i], &current_position );

				if ( mx_status.code != MXE_SUCCESS )
					return mx_status;

				break;

			default:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
					"Invalid motor subclass = %d", 
					motor->subclass);
				break;
			}

			/* Check for hardware limits. */

			mx_status = mx_motor_positive_limit_hit(
					motor_record[i], &positive_limit_hit);

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			mx_status = mx_motor_negative_limit_hit(
					motor_record[i], &negative_limit_hit);

			if ( mx_status.code != MXE_SUCCESS )
				return mx_status;

			/* Display the result. */

			if ( flags & MXF_MTR_SHOW_MOVE ) {
				switch ( motor->subclass ) {
				case MXC_MTR_STEPPER:
					fprintf( output,
						"%10.4f  %8ld ",
						current_position,
						current_steps );
					break;

				case MXC_MTR_ANALOG:
					fprintf( output,
						"%10.4f        -",
						current_position );
					break;
				}
			}
		}

		if ( flags & MXF_MTR_SHOW_MOVE ) {
			fprintf( output, "\n" );
		}

		mx_msleep(1000);

		if ( motors_busy == FALSE ) {
			exit_loop = TRUE;
		}

		/* Check to see if we need to abort the motor moves. */

		if ( (flags & MXF_MTR_IGNORE_KEYBOARD) == 0 ) {
			/* Did someone hit a key on the keyboard? */

			if ( mx_user_requested_interrupt() ) {

				abort_motors = TRUE;

				fprintf( output, "Keyboard interrupt.\n");
			}
		}

		if ( positive_limit_hit ) {
			abort_motors = TRUE;
			fprintf( output, "Positive hardware limit hit.\n");
		}

		if ( negative_limit_hit ) {
			abort_motors = TRUE;
			fprintf( output, "Negative hardware limit hit.\n");
		}

		if ( abort_motors ) {
			for ( i = 0; i < num_motors; i++ ) {
				mx_motor_soft_abort( motor_record[i] );
			}

			return_code = FAILURE;
			exit_loop = TRUE;
		}
	}

	if ( return_code == SUCCESS ) {
		return MX_SUCCESSFUL_RESULT;
	} else {
		return mx_error(MXE_INTERRUPTED, fname, "Motor move aborted.");
	}
}

